import os
import numpy as np
from .lensmodel_wrapper import run_lensmodel
from .lensmodel_handler import lensmodel_handler
from .utils import full_modeling_result,compare_dicts,write_pickle,read_pickle
#,get_paths_before #?
#run_lensmodel,save_dict_to_hdf5,full_modeling_result,read_hdf5,compare_dicts,write_pickle
from .lensmodel_result_handler import Result_Handler
from .mcmc_lensmodel import make_mcmc_lensmodel

#maybe use YAML 
def set_up_dir(path_for_models,dir_models,system_name):
    if not os.path.isdir(path_for_models):
        os.mkdir(path_for_models)
    if not os.path.isdir(os.path.join(path_for_models,dir_models)):
        os.mkdir(os.path.join(path_for_models,dir_models))
    if not os.path.isdir(os.path.join(path_for_models,dir_models,system_name)):
        os.mkdir(os.path.join(path_for_models,dir_models,system_name))
    return os.path.join(path_for_models,dir_models,system_name)
# To do change all the save and reading to pickle

def automatic_modeling(system,max_separation=0.01,path_for_models=None,dir_models=None,model_All=False,do_mcmc=False,max_lap=10,relative_flux_error=0.2,use_informed_flux=False,use_real_error_flux=False,do_model_of=None,**kwargs):
    #Check for existing results
    #path_model_result_csv = os.path.join(set_up_dir(system_name,path_for_models,dir_models),"final_result",f"models_{system_name}.csv")
    system_name = system.name.values[0]
    modeling_path = set_up_dir(path_for_models,dir_models,system_name)
    path_save = os.path.join(modeling_path,"model_result.pickle")
    lensmodel_system = lensmodel_handler(modeling_path,system)
    center_mass_error = 0.003
    astrometry_error = 0.003 
    can_end = False
    #quick solution
    #available_lens_models = lensmodel_system.mass_models.keys()
    if len(lensmodel_system.images) == 2:
        available_lens_models_to_model = ['SIS', 'SIE','SIS+shear']
    if len(lensmodel_system.images) > 2:
        available_lens_models_to_model = ['SIS+shear','SIS+shear','SIE','SIE+shear',"POW",'POW+shear']
    if system.total_lens.values[0]>1:
        available_lens_models_to_model = np.array([[model if r==1 else model+f"-{r}G" for r in range(1,system.total_lens.values[0]+1)] for model in available_lens_models_to_model]).ravel()
    if system.z_l.astype(float).values[0] > system.z_s.astype(float).values[0]:
            return print(f"Check redshift zl ({system.z_l.values[0]}) cant be bigger than zs ({system.z_s.values[0]})")
    if os.path.exists(path_save):
        #print(path_save)
        dic_lens_system=read_pickle(path_save)
        System_Class =  Result_Handler(dic_lens_system,max_separation=max_separation)#.get_best_model(,print_best=print_best)
        if System_Class.can_end and not model_All and not do_mcmc:
            print(f"The code found a best model with max(sep)< {max_separation} it is {System_Class.current_best_n_model}")
            return System_Class,modeling_path
        elif System_Class.can_end and do_mcmc:
            #lens_system_results = result_handler(dic_lens_system,max_separation=max_separation,look_for_best_model=True)
            #print(f"We will perform the mcmc using {lens_system_results.best_model_name} and it converge ={can_end}")
            print(f"The code found a best model with max(sep)< {max_separation} it is {System_Class.current_best_n_model}")
            System_Class = make_mcmc_lensmodel(System_Class,System_Class.current_best_n_model,modeling_path,rewrite=False)
            return System_Class,modeling_path
    available_lens_models = do_model_of or available_lens_models_to_model
    print("available_lens_models:",available_lens_models)
    for lap in range(10):
        if can_end and not model_All:
                    break
        for _,mass_model in enumerate(available_lens_models):
            if can_end and not model_All:
                    break
            alredy_model = False
            model_setup=lensmodel_system.writter_model_data(mass_distribution=mass_model,center_mass_error=center_mass_error,\
                                                            relative_flux_error=relative_flux_error,astrometry_error=astrometry_error,\
                                                                use_informed_flux=use_informed_flux,use_real_error_flux=use_real_error_flux)
            for key, value in kwargs.items():
                if key=="print_save" and value:
                    print(path_save)
            if os.path.exists(path_save):
                dic_lens_system=read_pickle(path_save)
                System_Class =  Result_Handler(dic_lens_system,max_separation=max_separation)
                if System_Class.can_end and not model_All:
                    break
                for key,items in dic_lens_system.items():
                    if compare_dicts(items["model_setup"]["model_setup"],model_setup):
                        alredy_model = True
                        print(f"lap {lap} Modeled {system_name} using {mass_model} center_mass_error {center_mass_error} relative_flux_error {relative_flux_error} astrometry_error {astrometry_error}")
            if not alredy_model:
                print(f"lap {lap} Modeling {system_name} using {mass_model} center_mass_error {center_mass_error} relative_flux_error {relative_flux_error} astrometry_error {astrometry_error}")
                run_lensmodel(modeling_path,model_setup['name_run'])
                current_model_result=full_modeling_result(modeling_path,model_setup) #local solution or dictionary with all the results from a model 
                if not os.path.exists(path_save) and "final_step" in current_model_result.keys():
                    write_pickle({"model_0":current_model_result},os.path.join(modeling_path,"model_result.pickle"))
                    dic_lens_system=read_pickle(path_save) #just it is done to be sure we are cheking well
                elif "final_step" in current_model_result.keys():
                    dic_lens_system=read_pickle(path_save)
                    n = max([int(i.replace("model_","")) for i in dic_lens_system.keys()])+1
                    dic_lens_system[f"model_{n}"] = current_model_result
                    write_pickle(dic_lens_system,os.path.join(modeling_path,"model_result.pickle"))
            System_Class =  Result_Handler(dic_lens_system,max_separation=max_separation)
            if System_Class.can_end and not model_All:
                print(f"The code found a best model with max(sep)< {max_separation} it is {System_Class.current_best_n_model}")
                break
        
        can_end = System_Class.can_end 
        
        if lap >=  max_lap:
            break
                #return dic_lens_system,modeling_path
        if lap==0:
            center_mass_error = 0.01
        if lap==1:
            center_mass_error = 0.01
            astrometry_error = 0.01 
        if lap==2:
            relative_flux_error = 0.5
            center_mass_error = 0.01 
            astrometry_error = 0.003
        if lap==3:
            relative_flux_error = 0.5
            center_mass_error = 0.01 
            astrometry_error = 0.01
        if lap == 4:
            print("Looks like any model works")
            break
    if not do_mcmc:
        return System_Class,modeling_path
    else:
        System_Class = make_mcmc_lensmodel(System_Class,System_Class.current_best_n_model,modeling_path,rewrite=False)
    return System_Class,modeling_path  


from astropy import units as u
from astropy import constants as const
from astropy.cosmology import FlatLambdaCDM
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
from lmfit.models import LinearModel
from lmfit import Model
import uncertainties
from uncertainties import unumpy,ufloat
from uncertainties.core import AffineScalarFunc
from .v_cmb import calculate_projections


class MicrolensingTimescale:
    #?
    #Not sure about the errors maybe can think about it again in the future 
    #cm_s = u.cm / u.s# cm/s
    # Global constants with default values
    G = 1.3271244e+26  # cm³/s²/solar mass
    c = 2.998e+10  # Speed of light in cm/s
    gamma = 0.55  # Growth index for LambdaCDM
    #sigma_pec(0)=439+/-69 (Hess & Kitaura, 2016)
    sigma_pec_0 = 439  
    dsigma_pec_0 = 69  # Uncertainty in peculiar velocity dispersion
    ######################
    M = 0.3  # Lens mass in solar masses
    p_dM = 0.1  # Percentage error in lens mass
    i = np.pi / 3  # Inclination angle in radians
    #################################
    v0 = 369.82  # CMB velocity in km/s
    dv0 = 0.11  # Uncertainty in CMB velocity
    ##################################
    z_pec,v_pec = np.array([0.,.5,1.0,2.0,3.0]), np.array([179.544800,162.386536,144.020065,114.992302,95.906265])
    fit_model = Model(LinearModel().func, independent_vars=['x'])
    model_v_pec = fit_model.fit(v_pec, x=z_pec)
    def __init__(self,cosmology=None,H=70,om=0.3):
        self.H = H
        self.om = om
        self.gamma = 0.55
        self.c = 2.998e+10  # Speed of light in cm/s
        self.G =  1.3271244e+26  # cm³/(s²*solar mass)
        self.cosmology = cosmology or FlatLambdaCDM(self.H,self.om)    
    # def sigma_pec(self, z):
    #     """
    #     Calculate the peculiar velocity dispersion at redshift z.
    #     """
    #     f = (unumpy.uarray(self.sigma_pec_0, self.dsigma_pec_0) * self.cosmology.Om(z)**self.gamma) / \
    #         (((1+z)**0.5) * self.cosmology.Om(0)**self.gamma)
        
    #     return self._convert_uncertainties(f)
    def enclosed_mass(self,zl,zs,teta_e,dteta_e,G=None,c=None,Da=None,Da_ls=None,return_all=False):
        G = G or self.G
        c = c or self.c
        Da = Da or self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
        Da_ls = Da_ls or self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        Re = unumpy.uarray(teta_e, dteta_e) * 4.84813681109536e-06 #* Da[0]
        M_e = (Re**2*Da[0] * Da[1] * c**2)/(4* G* Da_ls)
        if return_all:
            return np.concatenate([MicrolensingTimescale.to_return(M_e),MicrolensingTimescale.to_return(Re*Da[0]*3.2407792894443653e-22)],axis=1)
        return MicrolensingTimescale.to_return(M_e)
    def cosmological_(self,zl,zs,G=None,c=None,Da=None,Da_ls=None):
        #just get the ratio between the angular diameter distance
        G = G or self.G
        c = c or self.c
        Da = Da or self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
        Da_ls = Da_ls or self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        return (Da[0] * Da[1])/Da_ls
    def sigma_pec(self,z, gamma=None,sigma_pec_0=None,dsigma_pec_0=None):
        """
        for vp eq (9) from https://iopscience.iop.org/article/10.3847/0004-637X/832/1/46
        for f eq (8) from  https://ui.adsabs.harvard.edu/abs/2022MNRAS.509.2994A/abstract
        
        gamma -- Growth index, default is 0.55 for LambdaCDM
        return values of peculiar velocity in km/s
        235 evencio
        """
        gamma = gamma or self.gamma
        sigma_pec_0 = sigma_pec_0 or self.sigma_pec_0
        dsigma_pec_0 = dsigma_pec_0  or self.dsigma_pec_0
        f_z = self.cosmology.Om(z)**gamma
        f_0 = self.cosmology.Om(0)**gamma
        vp = (unumpy.uarray(sigma_pec_0, dsigma_pec_0) * f_z)/(((1+z)**0.5)* f_0)

        return MicrolensingTimescale.to_return(vp)
    def Re(self,zl,zs,M=None,c=c,G=G,p_dM=p_dM,Da=None,Da_ls=None):
        """Einstein radius of the lens 
        for Re eq 1 from https://iopscience.iop.org/article/10.1088/0004-637X/738/1/96
        """
        M = M or self.M
        p_dM = p_dM
        dM = p_dM * M
        #if zl>zs:
        #   return "zl should be minor than zs"
        #Estos podrian estar como variables globales ?
        if Da is None:
            Da = self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
        if Da_ls is None:
            Da_ls = self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        Re = Da[1]*(4* G *unumpy.uarray(M, dM) *Da_ls/ (Da[0] * Da[1] * c**2))**(0.5)
        return MicrolensingTimescale.to_return(Re)
        
    def Rs(self,z,m,dm,zpt,lamda,c=c,i=None,Da=None,**kwards):
        """Source radius 
        for Re eq 3 from https://iopscience.iop.org/article/10.1088/0004-637X/738/1/96
        and also re scale the radius to lamda 0.814 micro meter 
        """
        i = i or self.i
        h = self.H / 100
        r_h = self.c / self.cosmology.H(0).to("cm/cm s").value
        if Da is None:
            Da = self.cosmology.angular_diameter_distance(z).to(u.cm).value
        re_scale_factor = (0.814/(lamda/(1+z)))**(4/3)
        f= re_scale_factor * 3.4e15*(Da/r_h)*(lamda**(3/2))*((zpt/3631)**(0.5))*(10**(-0.2*(unumpy.uarray(m,dm)-19)))*(h**(-1))*(unumpy.sqrt(np.cos(i))**-1)
        return MicrolensingTimescale.to_return(f)
        #if isinstance(f,AffineScalarFunc):
         #   return np.array([f.nominal_value,f.s])
        #return np.array([[i.nominal_value,i.s] for i in f])
    
    def sigma_dv(self,zl,zs,teta_e,dteta_e,c=c,Da=None, Da_ls=None):#velocity dispersion of the stars in the lens galaxy
        """_summary_
        for sigma_dv eq 1 from https://www.aanda.org/articles/aa/abs/2008/02/aa7534-07/aa7534-07.html
        Args:
            zl (_type_): _description_
            zs (_type_): _description_
            teta_e (_type_): _description_ arcsec(")
            dteta_e (_type_): _description_ arcsec(")
            c (_type_, optional): _description_. Defaults to c.
            Da (_type_, optional): _description_. Defaults to None.
            Da_ls (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_ return velocidad en km/s
        """
        #
        teta_e = teta_e * 4.84814e-6 #to - rad
        dteta_e = dteta_e * 4.84814e-6 #to - rad
        if Da is None:
            Da = self.cosmology.angular_diameter_distance([zl]).to(u.cm).value
        if Da_ls is None:
            Da_ls = self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        
        f=1e-5*0.5*c*unumpy.sqrt(unumpy.uarray(teta_e, dteta_e)*Da/(np.pi*Da_ls))
        
        return MicrolensingTimescale.to_return(f)
    def v_proyected(self,ra,dec, v0=None, dv0=None,verbose=False):
        """_summary_
        for v_proyected eq 1 from https://www.aanda.org/articles/aa/abs/2008/02/aa7534-07/aa7534-07.html
        Args:
            ra (_type_): _description_
            dec (_type_): _description_
            v0 (_type_, optional): _description_. Defaults to None.
            dv0 (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        
        #value and error from https://arxiv.org/pdf/1807.06205
        v0 = v0 or self.v0
        dv0 = dv0 or self.dv0
        v_north,v_east = calculate_projections(ra,dec)
        V0 = unumpy.uarray(v0, dv0)
        f=1*unumpy.sqrt((v_north*V0)**2+(v_east*V0)**2)
        return MicrolensingTimescale.to_return(f)
    
    def Ve(self,zl,zs,ra,dec,teta_e,dteta_e,v0=None,dv0=None,gamma=gamma,sigma_pec_0=sigma_pec_0 \
        ,c=c,dsigma_pec_0=dsigma_pec_0,Da=None, Da_ls=None,return_all=True,verbose=False): # separation redshift lens redshift source peculiar velocity lens peculiar velocity source
        """_summary_
        folowing equation 8 from https://iopscience.iop.org/article/10.3847/0004-637X/832/1/46
        Args:
            zl (_type_): _description_
            zs (_type_): _description_
            ra (_type_): _description_
            dec (_type_): _description_
            teta_e (_type_): _description_
            dteta_e (_type_): _description_
            v0 (_type_, optional): _description_. Defaults to None.
            dv0 (_type_, optional): _description_. Defaults to None.
            gamma (_type_, optional): _description_. Defaults to gamma.
            sigma_pec_0 (_type_, optional): _description_. Defaults to sigma_pec_0
            c (_type_, optional): _description_. Defaults to c.
            dsigma_pec_0 (_type_, optional): _description_. Defaults to dsigma_pec_0.
            Da (_type_, optional): _description_. Defaults to None.
            Da_ls (_type_, optional): _description_. Defaults to None.
            return_all (bool, optional): _description_. Defaults to True.
            verbose (bool, optional): _description_. Defaults to False.

        Returns:
            _type_: _description_
        """
        gamma = gamma or self.gamma
        sigma_pec_0 = sigma_pec_0 or self.sigma_pec_0
        dsigma_pec_0 = dsigma_pec_0 or self.dsigma_pec_0
        v0 = v0 or self.v0
        dv0 = dv0 or self.dv0
        if Da is None:
            Da_l,Da_s = self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
        else:
            Da_l,Da_s = Da
        if Da_ls is None:
            Da_ls = self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        
        p_l=unumpy.uarray(*self.sigma_pec(zl,gamma=gamma,sigma_pec_0=sigma_pec_0,dsigma_pec_0=dsigma_pec_0).T)
        p_s=unumpy.uarray(*self.sigma_pec(zs,gamma=gamma,sigma_pec_0=sigma_pec_0,dsigma_pec_0=dsigma_pec_0).T)
        v_d=unumpy.uarray(*self.sigma_dv(zl,zs,teta_e,dteta_e,c=c,Da=Da_l, Da_ls=Da_ls).T)
        v_cmb = unumpy.uarray(*self.v_proyected(ra,dec,v0=v0,dv0=dv0).T)
        f=unumpy.sqrt(2*((p_l*Da_s/((1+zl)*(Da_l)))**2)+2*((p_s/(1+zs))**2)+((v_cmb*Da_ls/((1+zl)*Da_l))**2)+2*((v_d*Da_s/((1+zl)*Da_l))**2))
        if verbose:
            print(f"p_l:{p_l},p_s:{p_s},v_d:{v_d},v_cmb:{v_cmb}")
        return   [MicrolensingTimescale.to_return(i) for i in [f,v_cmb,v_d,p_s,p_l]]
        # if isinstance(f,AffineScalarFunc):
        #     return np.array([f.nominal_value,f.s])
        # elif len(f.shape)>1:
        #     return np.array([[i.nominal_value,i.s] for i in f[0].T])
        # elif len(f.shape)==1:
        #     return np.array([[i.nominal_value,i.s] for i in f])
        # else:
        #     f = f.reshape(1,)[0]
        #     return np.array([f.nominal_value,f.s])
    def time_scales(self,zl,zs,ra,dec,m,dm,zpt,lamda,teta_e,dteta_e,
                    v0=None,dv0=None,c=c,M=None,G=G,p_dM=None,i=None,
                    gamma=gamma,sigma_pec_0=None,dsigma_pec_0=None,return_all=False):
        """time scales in years
        zl,zs,m,dm,zpt,lamda,teta_e,dteta_e"""
        gamma = gamma or self.gamma
        sigma_pec_0 = sigma_pec_0 or self.sigma_pec_0
        dsigma_pec_0 = dsigma_pec_0 or self.dsigma_pec_0
        v0 = v0 or self.v0
        dv0 = dv0 or self.dv0
        p_dM = p_dM or self.p_dM
        M = M or self.M
        i = i or self.i
        Da = self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
        Da_ls = self.cosmology.angular_diameter_distance_z1z2(zl,zs).to(u.cm).value
        V =self.Ve(zl,zs,ra,dec,teta_e,dteta_e,v0=v0,dv0=dv0,c=c,Da=Da, Da_ls=Da_ls,\
            sigma_pec_0=sigma_pec_0,dsigma_pec_0=dsigma_pec_0)
        v = unumpy.uarray(*V[0].T*1e5)
        re = unumpy.uarray(*self.Re(zl,zs,M=M,c=c,G=G,p_dM=p_dM,Da=Da,Da_ls=Da_ls).T)
        rs= unumpy.uarray(*self.Rs(zs,m,dm,zpt,lamda,c=c,G=G,i=i,Da=Da[1]).T)
        te = 3.17098e-8 *(re/v)
        ts = 3.17098e-8 *(rs/v)
        if return_all:
            A = [np.squeeze(MicrolensingTimescale.to_return(i)) for i in [te,ts,rs,re]]+[np.squeeze(i) for i in V]
            if (A[0]).shape ==(2,):
                return np.concatenate(A)
            return  np.concatenate(A, axis=1)
        return MicrolensingTimescale.to_return(te),MicrolensingTimescale.to_return(ts)
        # if isinstance(te, AffineScalarFunc):
        #     tes = np.array([te.nominal_value,te.s]),np.array([ts.nominal_value,ts.s])
        # elif len(te.shape)>1:
        #     tes = np.array([[i.nominal_value,i.s] for i in te[0].T]),np.array([[i.nominal_value,i.s] for i in ts[0].T])
        # elif len(te.shape)==1:
        #     tes = np.array([[i.nominal_value,i.s] for i in te]),np.array([[i.nominal_value,i.s] for i in ts])
        # else:
        #     te = te.reshape(1,)[0]
        #     ts = ts.reshape(1,)[0]
        #     tes = np.array([te.nominal_value,te.s]),np.array([ts.nominal_value,ts.s])
        # if return_all:
        #     return tes,
        # return tes
    
        #return pandas zl,zs,m,dm,zpt,lambda,teta_e,dteta_e,vcmb,vcm,c,m,g,p_dm,i,p_l,ep_l,p_s,ep_s,v_d,ev_d,V,ve,rs,ers,re,ere,te,ete,ts,ets
    def get_supertabla(pandas:pd.DataFrame):
        return     
    def get_distance(self,zl,zs):
        return self.cosmology.angular_diameter_distance([zl,zs]).to(u.cm).value
    
    @staticmethod
    def to_return(f):
        if isinstance(f,AffineScalarFunc):
            f= np.array([f.nominal_value,f.s])
        elif len(f.shape)>1:
            f= np.array([[i.nominal_value,i.s] for i in f[0].T])
        elif len(f.shape)==1:
            f= np.array([[i.nominal_value,i.s] for i in f])
        else:
            f = f.reshape(1,)[0]
            f= np.array([f.nominal_value,f.s])
        return f


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os
import subprocess
import itertools

import stat
import shutil
from datetime import datetime
from ast import literal_eval


from astropy.cosmology import FlatLambdaCDM
from  lenstronomy.Util import param_util
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions
from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
Cosmo=FlatLambdaCDM(H0=67.36,Om0=0.264+0.051, Ob0=0.051,Neff=3.04)#lcdm no confiable del todo
Cpath=os.getcwd()+"/" 

def models_disp(imagenes,dif):
    #all use SIE
    Models=[]
    if len(imagenes)==1:
        Models=["SIS"]
    if len(imagenes)==2:
        Models=["SIS","SIE","SIS+shear"]
    if len(imagenes)==3:
        Models=["SIS","SIE","SIS+shear"]
    if len(imagenes)==4:
        Models=["SIE","SIE+shear"]
        if len(dif)>1:
            Models.append("SIE-2G")
    if len(imagenes)>=5 and len(dif)>1:
        Models.append("SIE-2G")
    return Models
def copy_file(pad):
    shutil.copyfile(pad+"/critT.dat",pad+"/result/critT.dat")
    shutil.copyfile(pad+"/gridT.dat",pad+"/result/gridT.dat")
    shutil.copyfile(pad+"/third.start",pad+"/result/third.start")

def canbemodeled(p):
    nombre_sistema = p.name.drop_duplicates().values[0]
    band=chooseband(p,print_bandas=False)
    imagenes=p[p['type']=="ima"][["RA","DEC","dDEC","dRA"]].astype(float) #informacion de las imagenes
    dif=p[p['type']=="dif"][[i for i in p[p['type']=="dif"].columns if i not in ["RA","DEC","dDEC","dRA"]]]
    Problem=[]
    tieneproblemas=False
    if len(imagenes)==0:
        problem="ceroimage"
        Problem.append(problem)
    if len(imagenes)==1:
        problem="ring"
        Problem.append(problem)
        print([nombre_sistema,problem])
    if len(imagenes)>5:
        problem="mas5i"
        Problem.append(problem)
        print([nombre_sistema,"mas5i"])
    if len(imagenes)==3:
        problem="triple"
        Problem.append(problem)
        print([nombre_sistema,"triple"])
    if band=="Noneband":
        problem="Noband"
        Problem.append(problem)
        #print([nombre_sistema,"Noband"])
    if len(dif)==0:
        problem="NoG"
        Problem.append(problem)
        #print([nombre_sistema,problem])
    if len(Problem)>0:
           tieneproblemas=True
    return tieneproblemas,np.array([nombre_sistema,Problem],dtype=object)

def sacador(i,L): #i==lista L array con los segmentos a sacar 
    sa=np.array([])
    for II in range(len(L)):
        if II==0:
            continue
        sa=np.append(sa,(i[int(L[II-1]):int(L[II])]).replace(" ", ""))
    return sa
            #print(list_of_lines[I][int(L[II-1]):int(L[II])])
def db(x,y): 
    #x:componente x del vector
    #y:componente y del vector
    D=np.array([])
    for i in range(len(x)):
        for ii in range(len(x)):
            d=np.sqrt((x[i]-x[ii])**2+(y[i]-y[ii])**2)
            if d not in D:
                D=np.append(D,d)
    return D

def correction(magnification,no_corregido): #correccion
    """magnification:magnificacion de la imagen
        no_corregido:flujo del sistema sin corregir"""
    #A=10**(-no_corregido/2.5) 
    #B=abs(magnification)
    return -2.5*np.log10((10**(-no_corregido/2.5))/abs(magnification))

def chooseband(B,print_bandas=True):
    try:
        B=B.loc[B.index.get_level_values('type')=="ima"]
    except:
        B=B.loc[B['type']=="ima"]
    Flujos=['g', 'r',"V","R","I","H",'i','z','Y','F160W','F814W','F475X',"G","5GHz","B","Kp"]
    F=[]
    B=B.fillna("-")
    for i in Flujos:
        if "-" in B[i].values:
            continue
        else:
            F.append(i)
    if print_bandas:
        print(f"Bandas disponibles {F}")
    if len(F)==0:
        f="Noneband"
        #print("Banda a utilizar "+f)
    if len(F)>0:
        if "F814W" in F:
            f="F814W"
            #print("Banda a utilizar "+f)
        elif 'i' in F:
            f='i'
        elif 'I' in F:
            f='I'
            #printabla_qso_lensingB.csvt("Banda a utilizar "+f)
        else:
            f=F[0]
            #print("Banda a utilizar "+f)
    return f
    
def getcurves(pad,crit="critT.dat",grid="gridT.dat"):
    files=os.listdir(pad)
    if crit in files and grid in files:
        files=[crit,grid]
        CG=[]
        for i in files:
            P=np.empty((0,4),float)#array vacia
            f=open(pad+"/"+i,"r")
            lines=f.readlines()
            for L in lines:
                if "Columns" in L or len(L)==1:
                    continue
                else:
                    p_critic=[]
                    for l in range(len(L)):
                        if " "==L[l] and " " !=L[l-1]:
                            p_critic.append(int(l))
                if len(p_critic)<4:
                    p_critic=[0]+p_critic
                p_critic=p_critic+[len(L)]
                p=np.array([float(L[p_critic[o]:p_critic[o+1]]) for o in range(len(p_critic)-1)])
                P=np.vstack([P,p])
            CG.append(P)
        curvas={"xc":CG[0][::,0],"yc":CG[0][::,1],"uc":CG[0][::,2],"vc":CG[0][::,3],
                "xg":CG[1][::,0],"yg":CG[1][::,1],"ug":CG[1][::,2],"vg":CG[1][::,3]}
    else:
        curvas="SAD"
        print("No estan los archivos")
    return curvas

def lensmodel(pad,nombre_run,work_pad=Cpath):
    """pad:directorio donde se quiere operar
    nombre de la run a aplicar"""
    x=pad+"/"+nombre_run
    preprocess_path =work_pad+"lensmodel"
    #print(preprocess_path)
    st = os.stat(preprocess_path)
    os.chmod(preprocess_path, st.st_mode | stat.S_IEXEC)   
    with open(x, 'r') as f:
        proc = subprocess.Popen([preprocess_path], stdin = f, stdout=subprocess.DEVNULL,
  stderr=subprocess.DEVNULL)
        proc.wait()
    for i in ['crit.dat','grid.dat','chitmp.dat','best.sm',"chitmp.dat"]:
        try:
            os.remove(os.path.join(Cpath,i))
            #print(f'The file {file_path} has been successfully removed.')
        except OSError as e:
            pass
    return

def getchis(imagenes,pad,archivo):
    """imagenes:solo para obtener el numero de imagenes del sistema
    pad:directorio donde se buscara el archivo
    archivo:nombre del archivo a utilizar
    """

    n_i=len(imagenes) # numero de imagenes 
    factorcorrection=1.143e-9
    completName = os.path.join(pad,archivo)
    file = open(completName, "r")
    list_of_lines = file.readlines()
    chix=np.array([])
    for I in range(len(list_of_lines)):
        L=np.array([])
        if 'CHISQ:'in list_of_lines[I]:
            for ii in range(len(list_of_lines[I])):
                if " "==list_of_lines[I][ii] and " "!=list_of_lines[I][ii-1]:
                    chix=np.append(chix,ii)
            chi_t=float(list_of_lines[I][int(chix[0]):int(chix[1])])
            chi_p=float(list_of_lines[I][int(chix[1]):int(chix[2])])
            chi_f=float(list_of_lines[I][int(chix[2]):int(chix[3])])
    chi={"chi_t":chi_t,"chi_p":chi_p,"chi_f":chi_f}
    return chi

def getresultados(i,pad,archivo): #nombre de una galaxia dentro de la lista de datos 
    """"i:nombre sistema
        pad:directorio
        archivo que contiene la informacion
     """
    completName = os.path.join(pad,archivo)
    file = open(completName, "r")
    list_of_lines = file.readlines()
    A=np.empty((0,10),float)#array vacia
    for I in range(len(list_of_lines)):
        L=np.array([])
        if 'alpha'in list_of_lines[I]:
            for ii, txt in enumerate(list_of_lines[I]):
                if txt==" " and list_of_lines[I][ii-1] !=" ":
                    L=np.append(L,ii)
            L=np.append(L,len(list_of_lines[I])-1)
            a=sacador(list_of_lines[I],L)[1:]
            A=np.vstack([A,a])
        if "posn:" in list_of_lines[I]:
            for ii, txt in enumerate(list_of_lines[I]):
                if txt==" " and list_of_lines[I][ii-1] !=" ":
                    L=np.append(L,ii)
            L=np.append(L,len(list_of_lines[I])-1)
            S=sacador(list_of_lines[I],L)[1:3]
    A=A.astype(float)   
    Datos={"s_x":float(S[0]),
           "s_y":float(S[1]),"p[1]":A[::,0],"p[2]":A[::,1],"p[3]":A[::,2],
           "p[4]":A[::,3],"p[5]":A[::,4],"p[6]":A[::,5],"p[7]":A[::,6],"p[8]":A[::,7],"p[9]":A[::,8],"p[10]":A[::,9]}
    return Datos
    #corrected=correction(1.143e-9,magnification,no_corregidos)

def getkg(pad,imagenes,band):
    for f in os.listdir(pad):
        if "kappa" in f:
            pad2=f
                #K.append(f)
        #k=dict.fromkeys([*K])
    f=open(pad+"/"+pad2,"r")
    lines=f.readlines()
    S=np.empty([0,6])
    for x in lines:
        if "output" in x or "source" in x:
            continue
        else:
            critic=[]
            for I in range(len(x)):
                if " "==x[I] and " "!=x[I-1]: 
                    critic.append(int(I))
                    if x[I+1]=="n":
                        print("hola")
            critic.append(len(x))
        L=[float(x[critic[i]:critic[i+1]]) for i in range(len(critic)-1)][1:6]
        L.append(1/((1-L[2])**2-L[3]**2))
        S=np.vstack((S,L))
    #correction(magnification,no_corregido)
    keys=["RA","DEC","kappa","gamma","theta","magnification"]
    k=dict.fromkeys(keys)
    for i,txt in enumerate(keys):
        k[txt]=S[::,i]
    dk=pd.DataFrame(k)
    dk["component"]=imagenes["component"].values
    dk["corrected"]=correction(dk["magnification"].values,imagenes[band].values)
    dk=dk.reindex(["component",*keys,"corrected"], axis=1)
    return dk

def dchi(nombre_sistema,pad,archivo,imagenes,dif,model,error_cm,error_flux,kapagamma=True,band="NOINFO",Psystem="NOINFO"): #nombre de una galaxia dentro de la lista de datos 
    """nombre_sistema:nombre galaxia
    pad: direccion en donde esta el archivo a extraer
    archivo:nombre del archivo
    imagenes: para obtener las componentes de cada imagen
    r_f: razon de flujo utilizada en el calculo
    e_d: error utilizado en el defractor
    er_f: error utilizado en la razon de flujos
    l:en esta version corresponde a el run a utlizar"""
    
    result=getresultados(nombre_sistema,pad,archivo)
    dis=np.empty([0,9])
    ################sacar_data_third.dat###############
    completName = os.path.join(pad,archivo)
    file = open(completName, "r")
    list_of_lines = file.readlines()
    nombres_imagenes=imagenes["component"].values
    n_i=len(nombres_imagenes)
    chi_t,chi_pos,chi_f=chi_t,chi_pos,chi_f=np.transpose(np.repeat(np.array(list(getchis(imagenes,pad,archivo).values()))[None,:],n_i, axis=0))
    fecha=str(datetime.now().year)+str(datetime.now().month)+str(datetime.now().day)+str(datetime.now().hour)+str(datetime.now().minute)
    for I in range(len(list_of_lines)):
        L=np.array([])
        if 'alpha'in list_of_lines[I]:
            for ii, txt in enumerate(list_of_lines[I]):
                if txt==" " and list_of_lines[I][ii-1] !=" ":
                    L=np.append(L,ii)
            L=np.append(L,len(list_of_lines[I])-1)
            A=sacador(list_of_lines[I],L)[1:]
        if "posn:" in list_of_lines[I]:
            for ii, txt in enumerate(list_of_lines[I]):
                if txt==" " and list_of_lines[I][ii-1] !=" ":
                    L=np.append(L,ii)
            L=np.append(L,len(list_of_lines[I])-1)
            S=sacador(list_of_lines[I],L)[1:3] #posicion de la fuente
        if ">" in list_of_lines[I]:
            for ii, txt in enumerate(list_of_lines[I]):
                if txt==" " and list_of_lines[I][ii-1] !=" ":
                    L=np.append(L,ii)
            L=np.append(L,len(list_of_lines[I])-1)
            X=sacador(list_of_lines[I],L)
            dis=np.vstack((dis,np.array([X[0].replace(" ", ""),X[1].replace(" ", ""),X[3],X[6],X[9],X[10],X[11],X[12],X[2].replace("(", "").replace(")", "")])))##reemplazar algo en un str
    
    A=A.astype(float)
    dis=dis.astype(float)
    
    d_im={"name":np.repeat(nombre_sistema,n_i),'image':nombres_imagenes,"RA":dis[:,0],'~RA':dis[:,4],
    "deltaRA":dis[:,0]-dis[:,4],"e_RA":dis[:,8],
    "DEC":dis[:,1],'~DEC':dis[:,5],"deltaDEC":dis[:,1]-dis[:,5],
    "e_DEC":dis[:,8],'F':np.abs(dis[:,2]),'~F':np.abs(dis[:,6]),"deltaF":dis[:,2]-dis[:,6],
    "error_flux":[error_flux for i in range(n_i)],"error_cm":[error_cm for i in range(n_i)],"chis_t":chi_t,"chi_f":chi_f,"chi_p":chi_pos,
    "fecha":np.repeat(fecha,n_i),"model":np.repeat(model,n_i),"banda":np.repeat(band,n_i),"Psystem":np.repeat(Psystem,n_i)}
    for i in result:
        if "s" not in i :#and len(result[i])>1
            d_im[i]=[[result[i][XD] for XD in range(len(result[i]))] for xd in range(n_i)]
        else:
            d_im[i]=np.repeat(result[i],n_i) 
    ###faltaria definir para en caso de tener mas de un G########
    for f in os.listdir(pad):#listdir(pad) para hacer lista de los objetos en un directorio
        if "kappa" in f:
            pad2=f
                #K.append(f)
        #k=dict.fromkeys([*K])
    f=open(pad+"/"+pad2,"r")
    lines=f.readlines()
    S=np.empty([0,6])
    for x in lines:
        if "output" in x or "source" in x:
            continue
        else:
            critic=[]
            for I in range(len(x)):
                if " "==x[I] and " "!=x[I-1]: 
                    critic.append(int(I))
                    if x[I+1]=="n":
                        print("hola")
            critic.append(len(x))
        L=[float(x[critic[i]:critic[i+1]]) for i in range(len(critic)-1)][1:6]
        L.append(1/((1-L[2])**2-L[3]**2))
        S=np.vstack((S,L))
    #correction(magnification,no_corregido)
    keys=["RA","DEC","kappa","gamma","theta","magnification"]
    k=dict.fromkeys(keys)
    for i,txt in enumerate(keys):
        k[txt]=S[::,i]
    dk=pd.DataFrame(k)
    
    dk["image"]=imagenes["component"].values
    dk["corrected"]=correction(dk["magnification"].values,imagenes[band].values)
    dk=dk.reindex(["image",*keys,"corrected"], axis=1)
    dk=dk.drop(['RA', 'DEC'], axis=1)
    #d_dif={"name":nombre_sistema,'image':"G","RA":0,'~RA':A[1],"deltaRA":0-A[1],"e_RA":0,
    #"DEC":0,'~DEC':A[2],"deltaDEC":0-A[2],"e_DEC":0,
    #'F':np.nan,'~F':np.nan,"deltaF":0,"e_F":0,"chis_t":chi_t[0],"chi_f":chi_f[0],"chi_p":chi_pos[0],
    #"fecha":fecha,"L":l}
    d=pd.DataFrame(d_im)
    Distancias=[np.sqrt(np.sum(np.power(i,2))) for i in d[["deltaRA","deltaDEC"]].values]
    dk["d_max"] = np.repeat(np.max(Distancias),len(d))
    C=pd.merge(d,dk, on ='image')
    #D = d.append(d_dif, ignore_index=True)
    dchi= pd.DataFrame(C)
    
    return dchi

def lenstronomy_obj(nombre_sistema,pad,archivo,imagenes,dif,pandas,zl,zs):
    if type(pandas["p[1]"].iloc[0])==str:
        d_r={i:pandas[i].iloc[[0]].values[0] for i in ["s_x","s_y"]}
        b={i:literal_eval(pandas[i].iloc[[0]].values[0]) for i in ["p[1]","p[2]","p[3]","p[4]","p[5]","p[6]","p[7]","p[8]","p[9]","p[10]"]}
        d_r.update(b)
    else:
        d_r={i:pandas[i].iloc[[0]].values[0] for i in ["s_x","s_y","p[1]","p[2]","p[3]","p[4]","p[5]","p[6]","p[7]","p[8]","p[9]","p[10]"]}
    #getresultados(nombre_sistema,pad,archivo)  
    kwargs_lens=[]
    lens_model_list=[]
    #print(d_r)
    for i in range(len(d_r["p[5]"])):
        model='PEMD'
        angulo=d_r["p[5]"][i]-90
        q=1-d_r["p[4]"][i]
        center_x,center_y=d_r["p[2]"][i],d_r["p[3]"][i]
        theta_E=np.sqrt((1+q**2)/(2*q))*d_r["p[1]"][i]#theta_E_gravlens
        e1,e2 = param_util.phi_q2_ellipticity((angulo)*np.pi/180.,q=q)
        #print(e1,e2,"exentricitis")
        kwargs_l={'theta_E': theta_E, 'e1': e1, 'e2': e2, 'center_x': center_x, 'center_y': center_y, 'gamma': 2.0}
        lens_model_list.append(model)
        kwargs_lens.append(kwargs_l)
        if len(imagenes)==4 and i==0 and d_r["p[7]"][i] !=0:
            model='SHEAR'
            angulo=d_r["p[7]"][i]+180
            gamma=d_r["p[6]"][i]
            gamma1,gamma2=param_util.shear_polar2cartesian(phi=(angulo)*np.pi/180., gamma=gamma)
            #e1,e2 = param_util.phi_q2_ellipticity((angulo)*np.pi/180.,q=q)
            kwargs_l = {'gamma1': gamma1, 'gamma2': gamma2}
            lens_model_list.append(model)
            kwargs_lens.append(kwargs_l)
    #lensModel=LensModel(lens_model_list=['PEMD'])
    source_x,source_y=d_r["s_x"],d_r["s_y"]
    # data imagen
    numPix = 500  #  cutout pixel size
    deltaPix = 0.05  #  pixel size in arcsec (area per pixel = deltaPix**2)
    ##################
    """print(source_x, source_y,"sx,sy modelo lente")
    print(lens_model_list)"""
    """print(kwargs_lens)
    print(lens_model_list)"""
    
    lens_model_class = LensModel(lens_model_list=lens_model_list, z_lens=zl, z_source=zs, cosmo=Cosmo) #primer cambio
    lensModel = LensModel(lens_model_list=lens_model_list)
    lensModelExtensions = LensModelExtensions(lensModel=lensModel)
    lensEquationSolver = LensEquationSolver(lens_model_class)
    xima, yima = lensEquationSolver.findBrightImage(source_x, source_y, kwargs_lens,
                                                        min_distance=deltaPix, search_window=numPix * deltaPix, numImages=len(imagenes))
    r_imagenes=[xima,yima]
    """DC=np.array([center_x,center_y])
    print("Diferencia entre centros",DC)
    if np.max(abs(DC))>0.05:
        print("Se aumento el error del centro")"""
    coord=imagenes[["RA","DEC"]].values-dif[["RA","DEC"]].values[0]
    d=np.array([])
    for i in range(len(xima)):
        D=np.power(np.power(coord[::,0]-xima[i],2)+np.power(coord[::,1]-yima[i],2),0.5)
        d=np.append(d,min(D))
    distancia_imagenes=d
    """print(d, "Distancia entre imagenes obtenidos e ingresadas")
    if np.max(abs(d))>0.05:
        print("Distancia maxima mayor a 0.05")
    Lista=[lensModel,kwargs_lens,source_x,source_y,imagenes,dif]"""
    return d_r,r_imagenes,lensModel,kwargs_lens,source_x,source_y,r_imagenes,distancia_imagenes

def graf_lensmodel(pandas,dif,imagenes,pad,NS=False,name_model=None,nombre_sistema=None):
    if type(pandas["p[1]"].iloc[0])==str:
        d_r={i:pandas[i].iloc[[0]].values[0] for i in ["s_x","s_y"]}
        b={i:literal_eval(pandas[i].iloc[[0]].values[0]) for i in ["p[1]","p[2]","p[3]","p[4]","p[5]","p[6]","p[7]","p[8]","p[9]","p[10]"]}
        d_r.update(b)
    else:
        d_r={i:pandas[i].iloc[[0]].values[0] for i in ["s_x","s_y","p[1]","p[2]","p[3]","p[4]","p[5]","p[6]","p[7]","p[8]","p[9]","p[10]"]}
    P1=[]
    T=[]
    for i in range(len(d_r["p[5]"])):
        angulo=d_r["p[5]"][i]-90
        q=1-d_r["p[4]"][i]
        theta_E=np.sqrt((1+q**2)/(2*q))*d_r["p[1]"][i]#theta_E_gravlens
        T.append(round(theta_E,4))
        P1.append(round(d_r["p[1]"][i],4))
    d2=np.power(np.power(pandas["RA"].values-pandas["~RA"].values,2)+np.power(pandas["DEC"].values-pandas["~DEC"].values,2),0.5)
    if NS==False:
        plt.subplots(1, 1, figsize=(10,10), sharex=False, sharey=False)
        plt.scatter(pandas["RA"].values,pandas["DEC"],c="g",marker="D",label="Posición observada imagenes")
        for ii in range(len(pandas["RA"])):
            plt.gca().text(pandas["RA"].values[ii]-0.02, pandas["DEC"].values[ii]-0.02, pandas["image"].values[ii], rotation=0,fontsize=20,zorder=10)
        plt.scatter(pandas["~RA"].values,pandas["~DEC"],marker="D", facecolors='none', edgecolors='r',label="Posición estimada imagenes")
        for i in range(len(d_r["p[5]"])):
            plt.scatter(dif["RA"].values[i]-dif["RA"].values[0],dif["DEC"].values[i]-dif["DEC"].values[0],c="r",label="Posición observada lente")
            plt.scatter(d_r["p[2]"][i],d_r["p[3]"][i],facecolors='none', edgecolors='k',label="Posición estimada de la lente")
        C=getcurves(pad)
        u=C["xc"]
        v=C["yc"]
        s=np.array([])

        for i in range(len(u)-1):
            D=np.power(np.power(u[i]-u[i+1],2)+np.power(v[i]-v[i+1],2),0.5)
            s=np.append(s,D)
        S=np.argmax(s)+1
        #plt.plot(C["xc"][S::],C["yc"][S::],label="criticaB")
        plt.plot(C["uc"][0:S],C["vc"][0:S], alpha=0.5,label="Curva caustica")
        plt.plot(C["xc"][0:S],C["yc"][0:S], alpha=0.5,label="Curva critica")
        left, width = .1, .0
        bottom, height = .1, .0
        right = left + width
        top = bottom + height
        ax = plt.gca()
        #p = plt.Rectangle((left, bottom), width, height, fill=False)
        #p.set_transform(ax.transAxes)
        #p.set_clip_on(False)
        #plt.gca().add_patch(p)
        #print(str(nombre_sistema)+" chi_t= "
         #               +str(pandas['chis_t'].values[0][0])+" model "+str(pandas['model'].values[0])
          #              + " p[1]="+str(P1)+" error_cm="+str(pandas["error_cm"].values[0][0]))
        print(f"{nombre_sistema} model={pandas['model'].values[0]} p[1]={P1}")
        print(f"chi_t = {pandas['chis_t'].values[0]}  error_cm= {pandas['error_cm'].values[0]} ")
        #print(f"chi_t={pandas['chis_t'].values[0]}, error_cm={} ")
        #pandas["error_cm"].values[0]
        print(f"separacion_maxima={np.max(d2)} y separacion_minima={np.min(d2)}") 
        
        
        # plt.gca().text(left, bottom, nombre_sistema+" chi_t= "
        #                +str(pandas['chis_t'].values[0])+" model "+str(pandas['model'].values[0])
        #                + " p[1]="+str(P1)+" error_cm="+str(pandas["error_cm"].values[0]),
        #         horizontalalignment='left',
        #         verticalalignment='top',
        #         transform=ax.transAxes)
        #plt.plot(C["uc"][S::],C["vc"][S::],label="causticaB")
        plt.scatter(pandas["s_x"],pandas["s_y"],c="y", alpha=1,marker="*",label="Posición estimada de la fuente")
        #print(d2)
        plt.ylabel(r'$\alpha$'+"[\"]", fontsize=20)
        plt.xlabel(r'$\delta$'+"[\"]", fontsize=20)
        plt.gca().invert_xaxis()
        plt.legend(bbox_to_anchor =(1.5,1))
        plt.gca().set_aspect('equal', adjustable='box')
        plt.savefig(pad+f"/model_{nombre_sistema}_{name_model}.png", bbox_inches='tight', dpi=400)
        plt.show()
    return d2
def Escritor(nombre_sistema,pad,imagenes,banda,dif,z_lens,z_source
             ,error_cm=[0.003],error_flux=0.2,Ngal=1,
             CM=False,noC=False,error_image=0.003,model="SIE"
             ,nombre_run="run_general",R_chi=None,use_real_error_flux=False):
    """"
    nombre_sistema
    pad: directorio donde se realizara todo el proceso
    imagenes: pandas_con informacion
    dif: pandas con informacion lentes
    error_cm list error asociado al centro de masa
    mG:bool si True se modelara las multiples galaxias del sistema lente
    CM: bool si es True se aumentara el error asociado a la galaxia lente
    Ngal: numero de galaxias que se ajustaran
    noC: bool si es True significa que se intentara modelar un sistema sin centro de masa conocido(work in progress)
    error_flux: error maximo en el flujo
    error_image: error en la pos de las imagenes
    """
    
    #calculate first approx p[1]
    alfa_0=np.array([])
    for i in range(len(imagenes)):
        ii=np.power(np.power(imagenes["RA"].values-imagenes["RA"].values[i],2)+np.power(imagenes["DEC"].values-imagenes["DEC"].values[i],2),0.5)
        alfa_0=np.append(alfa_0,ii/2) 
    alfa_0=round(np.max(alfa_0),2)
    #create de directory for the system an the results
    if os.path.isdir(pad)==False:
        os.mkdir(pad)
    if os.path.isdir(pad+"/result")==False:
        os.mkdir(pad+"/result")
    #########Writte data file###############
    name_of_file =str(nombre_sistema)+".dat"
    completeName = os.path.join(pad,name_of_file)
    f = open(completeName, 'w')
    #f.read
    f.write("#"+str(nombre_sistema)+'\n')
    ##########Numero de galaxias#########
    Ngal=1
    error_cm=error_cm#Error general
    if model=="SIE-2G":#multiples galaxias
        Ngal=len(dif)
        if len(error_cm)==1:
            error_cm=[0.05]*Ngal 
        if CM==True:
            error_cm=[0.0501]*Ngal
        print(error_cm)
    f.write(str(Ngal)+"  # number of lens galaxy"+'\n') # cambiar
    #################Error de G#####################
    if CM==True and  model!="SIE-2G":
        if len(imagenes)==2:
            x1,x2=imagenes["RA"].values-dif["RA"].values[0]
            y1,y2=imagenes["DEC"].values-dif["DEC"].values[0]
            p1=np.array([x1,y1])
            p2=np.array([x2,y2])
            p3=np.array([0,0])
            error_cm=[np.linalg.norm(np.cross(p2-p1, p1-p3))/np.linalg.norm(p2-p1)]
        if len(imagenes)>2:
        #   x,y=zip(*dif[["RA","DEC"]].values)#para asignar
            error_cm=[0.05] 
            if np.max(dif[["dRA","dDEC"]].values)>=0.003 and np.max(dif[["dRA","dDEC"]].values)<0.05:
                error_cm=[np.max(dif[["dRA","dDEC"]].values)]
            if np.max(dif[["dRA","dDEC"]].values)==0.003:
                error_cm=[0.003*2]
    if noC==True:#no centro ingresado
        error_cm=[1]
    for ii in range (Ngal):
        ra,dec=dif["RA"].values[ii]-dif["RA"].values[0],dif["DEC"].values[ii]-dif["DEC"].values[0]
        f.write(str(round(ra,3))+" "+str(round(dec,3))+" "+str(error_cm[ii])+" "+str(error_cm[ii])+" # position"+'\n') #arreg
        f.write("0.0 1000. # R_eff observed: 0.59 +/- 0.06"+'\n')
        f.write("0.0 1000. # PA unconstrained in observations"+'\n')
        f.write("0.0 1000. # observed e < 0.07 at 1sigma"+'\n')
    f.write('\n')
    f.write("1 # nºsource"+'\n')
    f.write(str(len(imagenes))+" # n° images of the source"+'\n')
    f.write("#flujo de "+str(banda)+'\n') 
    f.write("#Ra Dec flux sigmax sigmaflux tdel sigma(tdel) part "+'\n') 
    cord_imagenes=imagenes[["RA","DEC"]].values-dif[["RA","DEC"]].values[0]
    flujos=imagenes[banda].values
    error_cord=imagenes[["dRA","dDEC"]].values
    for ii, txt in enumerate(imagenes["component"].values): 
        a=cord_imagenes[:,0][ii]#ra
        b=cord_imagenes[:,1][ii] #Rdec
        c=((-1)**(ii))*10**(((-flujos[ii])+min(flujos))/2.5) #flujo divido
        if np.max(flujos)<=5:
            c=((1)**(ii))*flujos[ii]
        #r_f=np.append(r_f,abs(c))
        #d=eI[ii]
        #if len(error_cord[ii])>1:
        if np.max(flujos)>30:
            c=((-1)**(ii))*flujos[ii]/np.max(flujos)
        d=np.max(error_cord.flatten())
        d=error_image
        #np.sqrt(D[D["image"]!="G"]["dRA"].values[ii]**2+D[D["image"]!="G"]["dDEC"].values[ii]**2) #mejora el chi_p podria seguir poniendo 0 
        if use_real_error_flux==True:
            e=imagenes["d"+band].values[ii]
        g=0 #time delay
        h=1000 #deltatime delay
        f.write(str(round(a,10))+" "+str(round(b,10))+" "
                +str(round(c,10))+" "+str(round(d,10))+" "
                +str(round(e,10))+" "+str(round(g,10))+" "+str(round(h,10))+" "+str(txt)+'\n')
    f.close()
    """
    p[1] = M = mass scale
    (p[2],p[3]) = (x0,y0) = galaxy position
    (p[4],p[5]) = (e,θe) = ellipticity parameters
    (p[6],p[7]) = (γ,θγ) = external shear parameters
    (p[8],p[9]) = (s,a) = misc., often scale radii
     p[10] = α = misc., often a power law index"""
    alfa_0=np.array([])
    for i in range(len(imagenes)):
        ii=np.power(np.power(imagenes["RA"].values-imagenes["RA"].values[i],2)+np.power(imagenes["DEC"].values-imagenes["DEC"].values[i],2),0.5)
        alfa_0=np.append(alfa_0,ii/2) 
    alfa_0=round(np.max(alfa_0),2)
    ###creandocarpetas
    df=open(pad+"/"+nombre_run,'w')
    gf,sr="#set gridflag = 0 # don't need the tiling","#set restart = 2"
    chm="set chimode = 0 #sourceplane"
    if len(imagenes)==2:
        gf="#set gridflag = 0 # don't need the tiling"
        chm="set chimode = 0 #sourceplane"
    if len(imagenes)==4 and model=="SIE+shear":
        sr="set restart = 2"
    inicio=["#Modelado","set omitcore = 0.05","data "+pad+"/"+nombre_sistema+".dat #filewhitdata"
        ,"set omega = 0.3","set lambda = 0.7","set hval = 0.7"
        ,"set hvale = 1000  # Uncertainty in H_0 in units of 100 km/s/Mpc",chm
        ,"set zlens = "+str(z_lens)+"# lens redshift","set zsrc ="+ str(z_source)+"# lens redshift",
       "set checkparity = 0    # don't worry about parities",gf,sr,
       "#Start modeling","startup "+str(Ngal)+" 1 #startup Ngal 1"]
    third="third"#ultimo archivo para realizarle el mcmc
    if len(imagenes)==2:
        if model=="SIS":
            L3=["alpha "+str(alfa_0)+" 0 0 0 0.0 0 0 0 0 1","1 1 1 0 0 0 0 0 0 0"
                ,"optimize "+pad+"/third"]
            #L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0  #second","optimize "+pad+"/second"]
            #L3=["#thirdpart","setlens "+pad+"/first.start","changevary 1"," 1 1 1 0 0 0 0 0 0 0 # third","optimize "+pad+"/third"]
            L = itertools.chain(L3)
        if model=="SIE":
            L1=["alpha "+str(alfa_0)+" 0 0 0.1 10.0 0 0 0 0 1","1 0 0 1 1 0 0 0 0 0"
                ,"varyone 1 5 -90.0 90.0 19 "+pad+"/first"]
            #L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0  #second","optimize "+pad+"/second"]
            L3=["#thirdpart","setlens "+pad+"/first.start","changevary 1"," 1 0 0 1 1 0 0 0 0 0 # third","optimize "+pad+"/third"]
            L = itertools.chain(L1,L3)
        if model=="SIS+shear":
            L1=["alpha "+str(alfa_0)+" 0 0 0. 0.0 0,1 10 0 0 1","1 0 0 0 0 1 1 0 0 0"
                ,"varyone 1 7 -90.0 90.0 19 "+pad+"/first"]
            #L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0  #second","optimize "+pad+"/second"]
            L3=["#thirdpart","setlens "+pad+"/first.start","changevary 1"," 1 0 0 1 1 0 0 0 0 0 # third","optimize "+pad+"/third"]
            L = itertools.chain(L1,L3)
    if len(imagenes)>2: #always whit shear
        L1=["alpha "+str(alfa_0)+" 0 0 0.03 10.0 0 0.0 0 0 1","1 0 0 1 1 0 0 0 0 0"
            ,"varyone 1 5 -90.0 90.0 19 "+pad+"/first"]
        L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 0 0 1 1 0 0 0 0 0 #second","optimize "+pad+"/second"]

        L3=["#thirdpart","set chimode= 1","setlens "+pad+"/second.start","changevary 1"," 1 1 1 1 1 0 0 0 0 0 # third","optimize "+pad+"/third"]
        L = itertools.chain(L1,L2,L3)
        if model=="SIE+shear": #"set restart = 2",
            L1=["alpha "+str(alfa_0)+" 0 0 0.1 10.0 0.1 10.0 0 0 1","1 0 0 1 1 0 0 0 0 0"
            ,"varytwo 1 5 -90.0 90.0 19 1 7 -90.0 90.0 19 "+pad+"/first"]
            L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 1 1 0 0 0 0 0  #second","optimize "+pad+"/second"]
            L3=["#thirdpart","setlens "+pad+"/second.start","changevary 1"," 1 1 1 1 1 1 1 0 0 0 # third","optimize "+pad+"/third"]
            L = itertools.chain(L1,L2,L3)
        # if model=="POW" and len(imagenes)==4:
        #     L1=["alpha "+str(alfa_0)+" 0 0 0.03 10.0 0 0.0 0 0 1","1 0 0 1 1 0 0 0 0 0"
        #     ,"varyone 1 5 -90.0 90.0 19 "+pad+"/first"]
        #     L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 0 0 1 1 0 0 0 0 1 #second","optimize "+pad+"/second"]

        #     L3=["#thirdpart","set chimode= 1","setlens "+pad+"/second.start","changevary 1"," 1 1 1 1 1 0 0 0 0 1 # third","optimize "+pad+"/third"]
    
    if model=="SIE-2G":
        print("este ajuste solo esta habilitado para sistemas con 2G")
        ra,dec=dif["RA"].values-dif["RA"].values[0],dif["DEC"].values-dif["DEC"].values[0]
        L1=[]
        for i in range(len(ra)):
            L1.append("alpha "+str(alfa_0/2)+" "+str(ra[i])+" "+str(dec[i]) +" 0.1 10.0 0 0.0 0 0 1")
        for i in range(len(ra)):
            L1.append("1 0 0 1 1 0 0 0 0 0")
        L1.append("varytwo 1 5 -90.0 90.0 19 2 5 -90.0 90.0 19 "+pad+"/first")   
        L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0 #second","1 1 1 0 0 0 0 0 0 0","optimize "+pad+"/second"]
        L3=["#thirdpart","setlens "+pad+"/second.start","changevary 1"," 1 0 0 1 1 0 0 0 0 0 # third","1 0 0 1 1 0 0 0 0 0","optimize "+pad+"/third"]
        L = itertools.chain(L1,L2,L3)
    final=["set plotmode    = 2","kapgam 3 "+pad+"/kappa_gamma.dat","plotcrit "+pad+"/critT.dat",
          "plotgrid "+pad+"/gridT.dat",
           "calcRein 0 "+pad+"/RE.dat","quit"]
    if model=="SIE-2G" and CM==True:
        ra,dec=dif["RA"].values-dif["RA"].values[0],dif["DEC"].values-dif["DEC"].values[0]
        #values=R_chi[R_model["model"]=="SIE+shear"]
        L1=[]
        for i in range(len(ra)):
            if i==0:
                L1.append("alpha "+str(alfa_0)+" "+str(ra[i])+" "+str(dec[i]) +" 0.1 10.0 0 0.0 0 0 1")
            else:
                L1.append("alpha "+str(alfa_0)+" "+str(ra[i])+" "+str(dec[i]) +" 0.1 10.0 0 0.0 0 0 1")
        for i in range(len(ra)):
            L1.append("1 0 0 1 1 0 0 0 0 0")
        L1.append("varytwo 1 5 -90.0 90.0 19 2 5 -90.0 90.0 19 "+pad+"/first")   
        L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0 #second","1 1 1 0 0 0 0 0 0 0","optimize "+pad+"/second"]
        L3=["#thirdpart","setlens "+pad+"/second.start","changevary 1"," 1 1 1 1 1 0 0 0 0 0 # third","1 1 1 1 1 0 0 0 0 0","optimize "+pad+"/third"]
        L = itertools.chain(L1,L2,L3)
    if model=="SIE-2G" and CM==True and R_chi!=None and len(imagenes)>1000:
        ra,dec=dif["RA"].values-dif["RA"].values[0],dif["DEC"].values-dif["DEC"].values[0]
        values=R_chi[R_model["model"]=="SIE+shear"]
        L1=[]
        for i in range(len(ra)):
            L1.append("alpha "+str(alfa_0/2)+" "+str(ra[i])+" "+str(dec[i]) +" 0.1 10.0 0 0.0 0 0 1")
        for i in range(len(ra)):
            L1.append("1 0 0 1 1 0 0 0 0 0")
        L1.append("varytwo 1 5 -90.0 90.0 19 2 5 -90.0 90.0 19 "+pad+"/first")   
        L2=["#secondpart","setlens "+pad+"/first.start","changevary 1","1 1 1 0 0 0 0 0 0 0 #second","1 1 1 0 0 0 0 0 0 0","optimize "+pad+"/second"]
        L3=["#thirdpart","setlens "+pad+"/second.start","changevary 1"," 1 1 1 1 1 0 0 0 0 0 # third","1 1 1 1 1 0 0 0 0 0","optimize "+pad+"/third"]
        L = itertools.chain(L1,L2,L3)
    if nombre_run=="mcmc":    
        gf="set gridflag = 0 # don't need the tiling"
        chm="set chimode = 0 #sourceplane"
        sr="#set restart = 2"
        inicio=["#Modelado","set omitcore = 0.05","data "+pad+"/"+nombre_sistema+".dat #filewhitdata"
        ,"set omega = 0.3","set lambda = 0.7","set hval = 0.7"
        ,"set hvale = 1000  # Uncertainty in H_0 in units of 100 km/s/Mpc",chm
        ,"set zlens = "+str(z_lens)+"# lens redshift","set zsrc ="+ str(z_source)+"# lens redshift",
       "set checkparity = 0    # don't worry about parities",gf,sr]
        L=["setlens  "+pad+"/result/"+third+".start","MCMCset 2 Nchain 10 maxsteps 10000","MCMCrun "+pad+"/mcmc"]
        final=["quit"]
    #una nueva run ,"setlens  "+pad+"/"+third+".start","MCMCset 2 Nchain 10","MCMCrun "+pad+"/mcmc"
    for i in inicio:
        df.write(i)
        df.write('\n')
    for i in L:
        df.write(i)
        df.write('\n')
    for i in final:
        df.write(i)
        df.write('\n')
    df.close()
    if use_real_error_flux==True:
        error_flux=imagenes["d"+band].values
    return error_cm,error_flux

def mcmc(pad):
    if os.path.isfile(pad+"/"+"mcmc.dat")==False:
        print("realizando el MCMC (aqui esta lensmodel)")
        lensmodel(pad,nombre_run)
    if os.path.isfile(pad+"/"+"mcmc_r.csv")==False:
        print("creando CSV")
        import multiprocessing
        f=open(pad+"/mcmc.dat","r")
        lines=f.readlines()
        global alfas #para que solucione el error picke variable
        def alfas(l):
            A,B,C=None,None,None
            if "alpha" in l:
                #Parameters.append(l)
                critic=[]
                for i,text in enumerate(l):
                    if text==" " and l[i-1]!=" ": 
                        critic.append(int(i))
                critic.append(len(l))
                A=[l[critic[i]:critic[i+1]] for i in range(len(critic)-1)]
                #L=L.append("alfa")
            if "CHISQ:" in l:
                B=float(l.replace("CHISQ:",""))
            if "ptsrc" in l:
                critic=[]
                for i,text in enumerate(l):
                    if text==" " and l[i-1]!=" ": 
                        critic.append(int(i))
                critic.append(len(l))
                C=[l[critic[i]:critic[i+1]] for i in range(len(critic)-1)]
            return A,B,C
        ncpu = multiprocessing.cpu_count()
        #print(ncpu)
        if __name__ == "__main__":
            pool = multiprocessing.Pool(ncpu)
            out1, out2, out3 = zip(*pool.map(alfas, lines))
        Alfas,Chis,Ptsrc  = list(filter(None,out1)),list(filter(None,out2)),list(filter(None,out3))
        LP=np.array(Alfas)
        Chi=np.array(Chis)
        SP=np.array(Ptsrc)
        if len(LP)==2*len(Chi):
            LP=LP[::2]
        Dic={"Chi":0,"alpha":LP[::,0],"p[1]":0,"p[2]":0,"p[3]":0,"p[4]":0,"p[5]":0,"p[6]":0
        ,"p[7]":0,"p[8]":0,"p[9]":0,"p[10]":0,"ptsrc":SP[::,0],"s[1]":0,"s[2]":0,"s[3]":0,"s[4]":0,"s[5]":0,"s[6]":0,"s[7]":0}
        for i,txt in enumerate([*Dic]):
            if i==0:
                Dic[txt]=Chi
                continue
            if i>0 and i<12:
                if txt=="alpha":
                    continue
                Dic[txt]=LP[::,i-1].astype(float)
                limit=i
                continue
            else:
                if txt=="ptsrc":
                    continue
                i=i-limit
                Dic[txt]=SP[::,i]
        if len(Dic["alpha"])/len(Dic['s[1]']) >1:
            print("sistema con multiples galaxias lentes solo se guardara los parametros de la principal")
            F=int(len(Dic["alpha"])/len(Dic['s[1]']))
            for k in ['alpha', 'p[1]', 'p[2]', 'p[3]', 'p[4]', 'p[5]', 'p[6]', 'p[7]', 'p[8]', 'p[9]', 'p[10]']:
                Dic[k]=Dic[k][::F]
        Dic=pd.DataFrame(Dic)
        Dic.to_csv(pad+"/"+"mcmc_r.csv",index=False)
    return print("MCMC listo") 


import uncertainties
from uncertainties import unumpy,ufloat,umath
import pandas as pd
import numpy as np 

def _2_5_log10_and_error(args):
    #ima int, error imag int, ref int, error int
    value =  -2.5*umath.log10(ufloat(*args[0:2])/ufloat(*args[2:4]))
    return value.nominal_value,value.s

def ratio_cont_core(file,ref,img=None,save=False,obj_name=None):
    integrated = pd.read_csv(file)
    if isinstance(img,str):
        img = [img]
    elif img==None:
        img = [i for i in integrated.imagen.unique() if i !=ref]
    Ref = integrated.loc[(integrated["imagen"]==ref).values & ([bool("cont" not in i) for i in integrated["name"]])].copy()
    Ima = integrated.loc[np.array([bool(i in img) for i in integrated["imagen"].values]) & ([bool("cont" not in i) for i in integrated["name"]])].copy()
    Ref_c = integrated.loc[(integrated["imagen"]==ref).values & ([bool("cont" in i) for i in integrated["name"]])].copy()
    Ima_c = integrated.loc[np.array([bool(i in img) for i in integrated["imagen"].values]) & ([bool("cont" in i) for i in integrated["name"]])].copy()
    for n,ima in enumerate(img):
        I = Ima[Ima["imagen"]==ima].copy()
        I_c = Ima_c[Ima_c["imagen"]==ima].copy()
        Razon_l = pd.DataFrame([[ref,ima,I["name"].values[i],np.mean(I[["core_max","core_min"]].values[i]),*_2_5_log10_and_error((*I[['area_continuo','area_continuo_error']].values[i], 
        *Ref[['area_continuo','area_continuo_error']].values[i])),*_2_5_log10_and_error((*I[['core_line','area_continuo_error']].values[i], *Ref[['core_line','area_continuo_error']].values[i])),] for i in range(len(I))]
        ,columns=["ref","ima","line","wavelength","r_cont","e_r_cont","r_core","e_r_core"])
        #
        Razon_c = pd.DataFrame([[ref,ima,I_c["name"].values[i],np.mean(I_c[["core_max","core_min"]].values[i]),*_2_5_log10_and_error((*I_c[['area_continuo','area_continuo_error']].values[i], 
        *Ref_c[['area_continuo','area_continuo_error']].values[i])),0,0] for i in range(len(I_c))]
        ,columns=["ref","ima","line","wavelength","r_cont","e_r_cont","r_core","e_r_core"])
        Razon = pd.concat([Razon_c, Razon_l], ignore_index=True)
        Razon["dif"] =  Razon["r_cont"] - Razon["r_core"]
        if save==True:
            Razon.to_csv(f"Ratio_cont_core_{obj_name}_ref_{ref}_ima_{ima}.csv", index=False)
            print(f"Saved ref image {ref} and image {ima}")



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button,RangeSlider
from lmfit import Model
from scipy.integrate import quad
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import pandas as pd
import os
import multiprocessing
from functools import partial
from .utils import format_scientific
from screeninfo import get_monitors

module_dir = os.path.dirname(os.path.abspath(__file__))

# Get the screen resolution
monitor = get_monitors()[0]  # Assuming you want the primary monitor
screen_width = monitor.width
screen_height = monitor.height

# Calculate the aspect ratio
aspect_ratio = screen_width / screen_height


def linear_fit(x, slope, intercept):
        return slope * x + intercept

def bootstrap_single_iteration(x, y, _):
    pto_medio=np.min(x)+(np.max(x)-np.min(x))/2
    if type(x)==np.ndarray:
        x_L,y_L=x[x<pto_medio],y[x<pto_medio]
        x_R,y_R=x[x>pto_medio],y[x>pto_medio]
    else:
        x_L,y_L=x.values[x<pto_medio],y.values[x<pto_medio]
        x_R,y_R=x.values[x>pto_medio],y.values[x>pto_medio]
    # Resample the dataset with replacement
    resampled_indices_R = np.random.choice(len(x_R), len(x_R), replace=True)
    resampled_x_R = x_R[resampled_indices_R]
    resampled_y_R = y_R[resampled_indices_R]
   
    resampled_indices_L = np.random.choice(len(x_L), len(x_L), replace=True)

    resampled_x_L = x_L[resampled_indices_L]
    resampled_y_L = y_L[resampled_indices_L]
    
    model = Model(linear_fit)
    params = model.make_params(slope=1, intercept=0)
    resampled_x=np.concatenate((resampled_x_L, resampled_x_R))
    
    resampled_y=np.concatenate((resampled_y_L, resampled_y_R))
    result = model.fit(resampled_y, params, x=resampled_x)

    fitted_m = result.params["slope"].value
    fitted_b = result.params["intercept"].value

    
    x_min = min(resampled_x)
    x_max = max(resampled_x)

    area = (fitted_m * (x_max**2 / 2) + fitted_b * x_max) - (fitted_m * (x_min**2 / 2) + fitted_b * x_min)
    return fitted_m, fitted_b, area

def bootstrap_Area(x, y, n_bootstrap=5000):
    ncpu = multiprocessing.cpu_count()
    bootstrap_func = partial(bootstrap_single_iteration, x, y)

    with multiprocessing.Pool(ncpu) as pool:
        results = pool.map(bootstrap_func, range(n_bootstrap))

    slope_fit_estimates, intercept_estimates, area_estimates = zip(*results)
    mean_slope_fit_estimates,std_slope_fit_estimates = np.median(slope_fit_estimates),np.std(slope_fit_estimates)
    mean_intercept_estimates,std_intercept_estimates = np.median(intercept_estimates),np.std(intercept_estimates)
    mean_area,std_area = np.median(area_estimates),np.std(area_estimates)
    
    
    return mean_slope_fit_estimates,std_slope_fit_estimates,mean_intercept_estimates,std_intercept_estimates,mean_area,std_area, slope_fit_estimates, intercept_estimates, area_estimates
    # mean_area, std_area, area_estimates,Slope_fit,Intercept_fit
def update(val, y_noisy_list):
    
    global slider_barrier12,barrier_values,y_curve,y_noisy,x_between_barriers,y_noisy_between_barriers,range_values  #barrier_values, y_curve, y_noisy_between_barriers
    # Extract barrier positions from sliders
    x_barrier1, x_barrier2 = slider_barrier12.val[0], slider_barrier12.val[1]
    x_barrier3, x_barrier4 = slider_barrier34.val[0], slider_barrier34.val[1]
    x_barrier5, x_barrier6 = slider_barrier56.val[0], slider_barrier56.val[1]

    # Additional barrier positions
    x_slider_range, y_slider_range = slider_range.val, slider_range_y.val
    x_slider_range2, y_slider_range2 = slider_range2.val, slider_range_y2.val

    # Combine the data between barriers 1 and 2, and between barriers 3 and 4
    barrier_values = [x_barrier1, x_barrier2, x_barrier3, x_barrier4, x_barrier5, x_barrier6]
    range_values = [ *slider_range.val, *slider_range_y.val,*slider_range2.val, *slider_range_y2.val]
    x_between_barriers = np.concatenate((x[(x >= x_barrier1) & (x <= x_barrier2)], x[(x >= x_barrier3) & (x <= x_barrier4)]))
    
    ax1_1.clear()
    ax1_2.clear()
    for key, y_noisy in y_noisy_list.items():
        y_noisy_between_barriers = np.concatenate((y_noisy[(x >= x_barrier1) & (x <= x_barrier2)], y_noisy[(x >= x_barrier3) & (x <= x_barrier4)]))
        ax1_1.scatter(x_between_barriers, y_noisy_between_barriers, alpha=1, label=f'Data (between limits) imagen {key}', zorder=3) #pensar en colores
        ax1_1.plot(X,Y[key],label=f'spectrum imagen {key}', alpha=0.6, zorder=1)
        # Fit a single linear model to the data between the barriers
        linear_fit_model = Model(linear_fit)
        params = linear_fit_model.make_params(slope=1, intercept=0)
        result = linear_fit_model.fit(y_noisy_between_barriers, x=x_between_barriers, params=params)
        slope_fit, intercept_fit = result.params['slope'].value, result.params['intercept'].value
        ax1_1.plot(x_between_barriers, linear_fit(x_between_barriers, slope_fit, intercept_fit), label=f'Fitted Linear Function for {key}', color='r', zorder=4)
        
        # Calculate the area under the linear function between Barrier 1 and Barrier 4
        area, _ = quad(linear_fit, x_barrier1, x_barrier4, args=(slope_fit, intercept_fit))

        # Calculate the curve after subtracting the area obtained from the fit
        y_curve = y_noisy - linear_fit(x, slope_fit, intercept_fit)
        Y_curve = Y[key] - linear_fit(X, slope_fit, intercept_fit)
        suma = np.sum(y_curve[(x_barrier5 <= x) & (x_barrier6 >= x)])

        # Plot the curve resulting from subtracting the area
        
        ax1_2.plot(X, Y_curve, label=f'Line without continium {key}', color='grey',alpha=0.5)
        ax1_2.hlines(y=[0, 0], xmin=min(x), xmax=max(x), colors='k', linestyles='dashed', zorder=3)
        # Display the area as text inside the plot
    # Plot the fitted linear function   
    ax1_1.axvline(x=line_info["line_center"], color='k', label='Line',linestyle ="--")
    ax1_1.fill_betweenx([-100,100], x_barrier1, x_barrier2,
                        label='left continium', color='purple', alpha=0.2)

    # Create a region between x_barrier3 and x_barrier4
    ax1_1.fill_betweenx([-100,100], x_barrier3, x_barrier4,
                        label='right continium', color='g', alpha=0.2)

# Rest of the code remains the same up to this point

    ax1_1.legend(fontsize='xx-small', frameon=False,ncol=3)
    ax1_1.set_xlim(x_slider_range[0], x_slider_range[1])
    ax1_1.set_ylim(y_slider_range[0], y_slider_range[1])
    ax1_1.set_xlabel('wavelength')
    ax1_1.set_ylabel('Flux')
    
    ax1_2.fill_between(x[(x_barrier5 <= x) & (x_barrier6 >= x)], 
                   -100,100, 
                   label='line core region', color='r', alpha=0.3, zorder=2)
    #ax1_2.axvline(x=x_barrier5, color='m', linestyle='--', label='Barrier 5')
    #ax1_2.axvline(x=x_barrier6, color='c', linestyle='--', label='Barrier 6')
    ax1_2.axvline(x=line_info["line_center"], color='k', label='Line',linestyle ="--")
    ax1_2.set_xlim(x_slider_range2[0], x_slider_range2[1])
    ax1_2.set_ylim(y_slider_range2[0], y_slider_range2[1])
    ax1_2.set_xlabel('wavelenght')
    ax1_2.set_ylabel('Flux')
    ax1_2.legend(fontsize='small', frameon=False,ncol=3)
    
    plt.draw()

def equalility(panda_new,row):
    are_equal,fit_equal,barrier_equal=[],[],"?"
    tolerance = 1e-10
    for i in panda_new.columns:
        is_equal=row[i].values==panda_new[i].values
        if not row[i].values==panda_new[i].values and "y" not in i:
            is_equal = np.abs(row[i].values - panda_new[i].values) < tolerance
        are_equal.append(is_equal)
        if "right" in i or "left" in i:
            fit_equal.append(is_equal)
    return are_equal,fit_equal    
 
def save_data2():
    #global panda_new,panda_read,row,barrier_values,range_values,y_noisy_list,x_between_barriers,y_noisy_between_barriers,line_name,imagen,banda
    for imagen, y_noisy in y_noisy_list.items():
        #falta metodo de error
        panda_new = pd.DataFrame({
            "name":[line_name],
            "imagen":[imagen],
            # The code is checking if all elements in the `are_equal` list are not equal. If any element
            # is not equal to the others, the condition will be true.
            "band":[band],
            'min_x': [np.min(x)],
            'max_x': [np.max(x)],
            'left_min': [barrier_values[0]],
            'left_max': [barrier_values[1]],
            'right_min': [barrier_values[2]],
            'right_max': [barrier_values[3]],
            'core_min': [barrier_values[4]],
            'core_max': [barrier_values[5]],
            "wavelength":[(barrier_values[4]+barrier_values[5])/2],
            "x_range_min" :[range_values[0]],
            "x_range_max" :[range_values[1]],
            "y_range_min" :[range_values[2]],
            "y_range_max" :[range_values[3]],
            "x_range2_min" :[range_values[4]],
            "x_range2_max" :[range_values[5]],
            "y_range2_min" :[range_values[6]],
            "y_range2_max" :[range_values[7]],
        })

        are_equal,fit_equal=[False],[False]
        if name_file in os.listdir(os.getcwd()):
            panda_read = pd.read_csv(name_file)
            if any((panda_read[["name", "imagen"]].values == [line_name, imagen]).all(axis=1)):
                row=panda_read.loc[((panda_read['name'] == line_name) & (panda_read["imagen"] == imagen))]
                are_equal,fit_equal = equalility(panda_new,row)
        if not all(are_equal):
            panda_new['y_noisy'] = str(y_noisy.tolist())
            if not all(fit_equal):
                print(f"making the bootstraping for {line_name}, {imagen}")
                y_noisy_between_barriers = np.concatenate((y_noisy[(x >= barrier_values[0]) & (x <= barrier_values[1])], y_noisy[(x >= barrier_values[2]) & (x <= barrier_values[3])]))
                mean_slope_fit_estimates,_,mean_intercept_estimates,_,mean_area,std_area, _, _, _ = bootstrap_Area(x_between_barriers,y_noisy_between_barriers, n_bootstrap=5000)
                ycurve=y_noisy-linear_fit(x, mean_slope_fit_estimates,  mean_intercept_estimates)
                panda_new['y_curve'] = str(list(ycurve))
                #print(mean_slope_fit_estimates,mean_intercept_estimates,mean_area,std_area )
                panda_new["slope_fit"]= mean_slope_fit_estimates
                panda_new["intercept_fit"]=mean_intercept_estimates
                panda_new["area_continuo"]=mean_area
                panda_new["area_continuo_error"]=std_area
            else:
                for i in row.columns:
                    if i not in panda_new.columns:
                        panda_new[i]=row[i].values
                ycurve=eval(row['y_curve'].values[0])
            #aqui mal #print(panda_new["core_line"])
            if "cont" in line_name:
                panda_new["core_line"] = 0
            else:   
                panda_new["core_line"]=np.sum(np.array(ycurve)[(barrier_values[4]<=x) & (barrier_values[5]>=x)])
            try:
                if len(panda_read)>0:
                    if any((panda_read[["name", "imagen"]].values == [line_name, imagen]).all(axis=1)):
                        print(f'Data updated in {name_file} for {line_name}, {imagen}')
                        condition = ((panda_read['name'] == line_name) & (panda_read["imagen"] == imagen))
                        row_to_drop = panda_read.loc[condition]
                        panda_read = panda_read.drop(row_to_drop.index)
                        pd.concat([panda_read, panda_new],axis=0).to_csv(name_file, index=False)
                    else:
                        print(f"New row {line_name}, {imagen} in file {name_file}")
                        pd.concat([panda_read, panda_new],axis=0).to_csv(name_file, index=False)
            except:
                    print(f"New row {line_name}, {imagen}")
                    print(f"Creating file {name_file}")
                    panda_new.to_csv(name_file, index=False)
        else:
            print(f"No changes detected in row {line_name}, {imagen}")

def savesubplot(ax,bbox_inches,path):
    x1, x2 = ax.get_xlim()
    y1, y2 = ax.get_ylim()
    fig_save, ax_save = plt.subplots()
    ax_save.set_xlim(x1, x2)
    ax_save.set_ylim(y1, y2)
    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(path, bbox_inches=extent.expanded(*bbox_inches))
    plt.close(fig_save)

def on_save_button_clicked(event):
    #global y_curve,y_noisy,S,x_between_barriers,y_noisy_between_barriers,line_name,imagen,banda
    global images,panda_new,panda_read,row,barrier_values,range_values,y_curve,y_noisy,x_between_barriers,y_noisy_between_barriers,line_name,imagen,band,y_noisy_list
    save_data2()
    #save_data()   
    button_save.label.set_visible(True)
    fig.canvas.draw_idle()
    try:
        os.mkdir("images")
    except:
        pass
    savesubplot(ax1_2,(1.15, 1.25),f'images/fit_{line_name}_image_{"".join(list(Y.keys()))}_WC.png')
    savesubplot(ax1_1,(1.15, 1.3),f'images/fit_{line_name}_image_{"".join(list(Y.keys()))}.png')
    print("Image saved")
    
def on_remove_line_clicked(event):
    panda_previous=pd.read_csv(name_file)
    panda_previous = panda_previous[panda_previous["name"] != line_name]
    panda_previous.to_csv(name_file)
    print(f"deleted {line_name}")

def on_close_all(event):
    global stop_flag
    stop_flag = True  # Set the flag to True to stop the loop
    print("Requested to stop plotting.")

#continium lines should be have a more wide range to select windows 
#maybe add a function to save a restframe like file in cases were i found nice windows to use for other examples
global images

stop_flag=False

def window_analisis(pandas_spectra,name_obj,flux_column="flux_",lens_images=None,bands=None,use_preselected_lines=False,file=None):
    #obj= imagenes 
    #name_obj str
    zs = pandas_spectra.zs.values[0]
    global slider_barrier12,slider_barrier34,slider_barrier56,slider_range,slider_range_y,slider_range2,slider_range_y2,x,ax1_1,ax1_2,X,Y,line_info,y_noisy_list,line_name,band,name_file,button_save,fig
    window=pd.read_csv(os.path.join(module_dir,"rest_frame_windows.csv"))
    w = [[list(np.array(eval(ii.replace(" ",",").replace("nan","np.nan")))*(1+zs)) for ii in window[i].values] for i in window.columns if "line_name" not in i]
    #window[[i for i in window.columns if "line_name" not in i]] = [[np.array(eval(ii.replace(" ",",").replace("nan","np.nan")))*(1+zs) for ii in window[i].values] for i in window.columns if "line_name" not in i]
    window["left_range"] =w[0]
    window['ring_range'] =w[1]
    window['core_range'] =w[2]
    ##,barrier_values,y_curve,y_noisy,x_between_barriers,y_noisy_between_barriers,range_values
    if name_file==None: 
        name_file=f"Flux_cont_core_{name_obj}.csv"
    else:
        name_file = file
    if bands==None:
        bands = pandas_spectra.band.unique()
    for band in bands:
        spectra_band = pandas_spectra[pandas_spectra.band==band]
        spectra_band = spectra_band.dropna(axis=1, how='all')
        if lens_images==None:
            images = [i.replace(flux_column,"") for i in spectra_band.columns if flux_column in i and "G" not in i]
        
        for r in range(len(window)):
            
            line_info=window.iloc[r].copy()
            line_name=line_info["line_name"]
            if use_preselected_lines==True:
                if name_file in os.listdir(os.getcwd()):
                    panda_read=pd.read_csv(name_file)
                    if not any((panda_read[["name"]].values == [line_name]).all(axis=1)):
                        continue
            #maybe condition
            if "core_range" in line_info.keys() and "line_center" not in line_info.keys():
                mean_rango = np.mean(eval(str(line_info["core_range"])))
                line_info["line_center"] = mean_rango 
                rango=[mean_rango-1000,mean_rango+1000]
            else:
                rango=[line_info["line_center"]-1000,line_info["line_center"]+1000]
            if mean_rango<spectra_band["wavelength"].values[0] or mean_rango>spectra_band["wavelength"].values[-1]:
                continue 
            #######################################
            X,Y = spectra_band["wavelength"].values,{imagen:spectra_band[f"{flux_column}{imagen}"].values for imagen in images}
            x=spectra_band["wavelength"].values[(rango[0]<=spectra_band["wavelength"].values) & (rango[1]>=spectra_band["wavelength"].values)]
            y_noisy_list= {imagen:spectra_band[f"{flux_column}{imagen}"].values[(rango[0]<=spectra_band["wavelength"].values) & (rango[1]>=spectra_band["wavelength"].values)] for imagen in images}
            y_list_D=list(y_noisy_list.values())
            y_up,y_down=np.max(y_list_D)*1.1,np.min(y_list_D)
            q2 = np.percentile(y_list_D, 89)
            q3 = np.percentile(y_list_D, 95) 
            fig = plt.figure(figsize=(15, 15/ aspect_ratio))
            grid = plt.GridSpec(2, 2, width_ratios=[2, 2], height_ratios=[3, 1], hspace=0.4)
            ax1_1 = plt.subplot(grid[0, 0])
            ax1_2 = plt.subplot(grid[0, 1])
            
            
            bbox = ax1_1.get_position()
            bbox2 = ax1_2.get_position()
            
            # Add sliders for the barrier positions (as a column inside [0, 1])
            slider_barrier1_ax = plt.axes([bbox.x0, bbox.y0-0.2, bbox.width, 0.03]) #left conti
            slider_barrier4_ax = plt.axes([bbox.x0, bbox.y0-0.25, bbox.width, 0.03]) #right conti
            slider_range_ax = plt.axes([bbox.x0, bbox.y0-0.1, bbox.width, 0.03]) #range wavelenght 
            
            
            slider_range_ax2 = plt.axes([bbox2.x0, bbox2.y0-0.1, bbox2.width, 0.03]) #range wavelenght for core
            slider_barrier6_ax = plt.axes([bbox2.x0, bbox2.y0-0.225, bbox2.width, 0.03]) # line core
            
            y_slider_range_ay2=plt.axes([bbox2.x0+bbox.width+0.05,bbox2.y0, 0.01, bbox.height]) #range flux core
            y_slider_range_ay=plt.axes([bbox.x0-0.07, bbox.y0, 0.01, bbox2.height]) #range flux  
            
            
            
            if band=="NIR":
                y_up = q2*4
            L_valini=[line_info["line_center"]-100,line_info["line_center"]-50,line_info["line_center"]+50,line_info["line_center"]+100,line_info["line_center"]-10,line_info["line_center"]+10]
            range_values_O = [line_info["line_center"]-500, line_info["line_center"]+500,-y_up,y_up,line_info["line_center"]-100,line_info["line_center"]+100,-y_up,y_up]
            if name_file in os.listdir(os.getcwd()):
                panda_read=pd.read_csv(name_file)
                if any((panda_read[["name"]].values == [line_name]).all(axis=1)):
                    row=panda_read.loc[panda_read["name"]==line_name]
                    L_valini =[row[i].values[0] for i in row.columns if "left" in i] + [row[i].values[0] for i in row.columns if "right" in i] +[row[i].values[0] for i in row.columns if "core" in i]
                    try:
                        range_values_O = [row[i].values[0] for i in row.columns if "x_range_" in i]+[row[i].values[0] for i in row.columns if "y_range_" in i]+[row[i].values[0] for i in row.columns if "x_range2_" in i]+[row[i].values[0] for i in row.columns if "y_range2_" in i]
                    except:
                        pass
            ###########################
            slider_barrier12 = RangeSlider(slider_barrier1_ax, "left \ncontinium",np.min(x),line_info["line_center"],valinit=[L_valini[0], L_valini[1]],color="purple",alpha=0.5) 
            slider_barrier34 = RangeSlider(slider_barrier4_ax, "right \ncontinium",line_info["line_center"],np.max(x),valinit=[L_valini[2], L_valini[3]],color="green",alpha=0.2) 
            if band=="NIR":
                slider_barrier56 = RangeSlider(slider_barrier6_ax , "line core",line_info["line_center"]-100,line_info["line_center"]+100,valinit=[L_valini[4], L_valini[5]],color="r",alpha=0.2)
            elif "cont" in line_name:
                slider_barrier56 = RangeSlider(slider_barrier6_ax , "line core",line_info["line_center"]-1,line_info["line_center"]+1,valinit=[L_valini[4], L_valini[5]],color="r",alpha=0.2)
            else:
                slider_barrier56 = RangeSlider(slider_barrier6_ax , "line core",line_info["line_center"]-100,line_info["line_center"]+100,valinit=[L_valini[4], L_valini[5]],color="r",alpha=0.2)
            ###################################
            slider_range = RangeSlider(slider_range_ax, "Range",np.min(X),     np.max(X),valinit=range_values_O[0:2])
            slider_range_y=RangeSlider(y_slider_range_ay, "Range",-y_up,y_up,orientation="vertical",valinit=range_values_O[2:4], valfmt="")
            slider_range2 = RangeSlider(slider_range_ax2, "Range",np.min(x), np.max(x),valinit=range_values_O[4:6])
            slider_range_y2=RangeSlider(y_slider_range_ay2, "Range",-y_up,y_up,orientation="vertical",valinit=range_values_O[6:8], valfmt="")
            #slider_range_y2.set_visible(False) 
            slider_range_y2.valtext.set_visible(False)
            slider_range_y.valtext.set_visible(False)
            slider_range.valtext.set_visible(False)
            slider_range2.valtext.set_visible(False)
            
            ax_save_button = plt.axes([0.4, 0.02, 0.2, 0.04])
            button_save = Button(ax_save_button, 'Save', color='lightgoldenrodyellow', hovercolor='0.975')
            if name_file in os.listdir(os.getcwd()):
                if any((panda_read[["name"]].values == [line_name]).all(axis=1)):
                    ax_save_button = plt.axes([0.6, 0.02, 0.2, 0.04])
                    button_remove = Button(ax_save_button, 'remove line', color='lightgoldenrodyellow', hovercolor='0.975')
                    button_remove.on_clicked(on_remove_line_clicked)
            # Update the plot when slider values change
            ax_save_button = plt.axes([0.01, 0.95, 0.2, 0.04])
            button_close = Button(ax_save_button, 'Close', color='lightgoldenrodyellow', hovercolor='0.975')
            button_close.on_clicked(on_close_all)
            
            slider_barrier12.on_changed(lambda val: update(val, y_noisy_list))
            slider_barrier34.on_changed(lambda val: update(val, y_noisy_list))
            slider_barrier56.on_changed(lambda val: update(val, y_noisy_list))
            slider_barrier56.on_changed(lambda val: update(val, y_noisy_list))


            slider_range.on_changed(lambda val: update(val, y_noisy_list))
            slider_range_y.on_changed(lambda val: update(val, y_noisy_list))
            slider_range2.on_changed(lambda val: update(val, y_noisy_list))
            slider_range_y2.on_changed(lambda val: update(val, y_noisy_list))
            #slider_barrier8.on_changed(update)
            # Set up the event handler for the button

            button_save.on_clicked(on_save_button_clicked)
            
            ax1_1.set_xlim(np.min(x), np.max(x))
            ax1_1.set_xlabel('wavelength')
            ax1_1.set_ylabel('Flux')
            update(0.01, y_noisy_list)
            titulo = " and ".join(images)
            if len(images)>2:
                titulo = ",".join(images)
            fig.suptitle(f"Fit {line_name} line for image {titulo}", fontsize=16)
            #plt.gca().set_aspect('equal', adjustable='box')
            if stop_flag:
              #  plt.close()
                return
            #else:
            plt.show()
        images=None
            



import pytest


# a
@pytest.fixture
def get_main_arguments():
    return type('A', (), {'command_line_script_args': None, 'main_function_kwargs': {}})


@pytest.mark.runner_setup()
def test_cli(
    get_main_arguments,
    isolated_cli_runner,
):
    from qumas.cli import main

    main_arguments = get_main_arguments()
    result = isolated_cli_runner.invoke(
        main,
        args=main_arguments.command_line_script_args,
        input=None,
        env=None,
        catch_exceptions=False,
        **main_arguments.main_function_kwargs,
    )
    assert result.exit_code == 0
    assert result.stdout == ''



import sys


def test_invoking_cli_as_python_module(run_subprocess):
    result = run_subprocess(
        sys.executable,
        '-m',
        'qumas',
        '--help',
    )
    assert result.exit_code == 0
    assert result.stderr == ''
    assert result.stdout.split('\n')[0] == "Usage: qumas [OPTIONS]"



from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np
from matplotlib.patches import Polygon

def map_plot(mag_map_2d,cmap='RdYlBu',vmin=-1.5,vmax=0.5,cbar_side="right",remove_labels=True,cbar_label=r"$-2.5 \, \log(\mu/\langle \mu \rangle)$",save="",**kwargs):
    
    text = kwargs.get("text")
    fig, ax = plt.subplots(1, 1, figsize=(20*1, 10*1))

    im = ax.imshow(mag_map_2d,cmap=cmap, extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    
    #im.cmap.reversed()
    #divider = make_axes_locatable(ax)
    #side = "right"
    #cax = divider.append_axes(side, size="5%", pad=0.2)  # Adjust size and padding
    
    #ax.set_title(titles[i], fontsize=30)
    cbar = plt.colorbar(im)
    
    #x_min, x_max, y_min, y_max = im.get_extent()
    cbar.ax.set_position([0.33, 0.15, 0.05, 0.73])
    if cbar_side=="right":
        cbar.ax.set_position([0.75, 0.15, 0.05, 0.73]) 
        
    cbar.ax.tick_params(labelsize=30)
    cbar.set_label(cbar_label, fontsize=30,position=cbar_side)
    cbar.ax.yaxis.set_label_position(cbar_side)
    cbar.ax.yaxis.set_ticks_position(cbar_side)
        # Adjust tight_layout to leave space for the suptitle
    if text:
        ax.text(0.05, 0.95, text, transform=ax.transAxes,
             fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))    
    #height, width = mag_map_2d.shape

    bluest_color = im.cmap(im.norm(vmin))  # Normalize and map vmin to color
    #if side=="right":
    trixy = np.array([[0, 0], [1, 0], [0.5, -0.05]])
    p = Polygon(trixy, transform=cbar.ax.transAxes, 
                        clip_on=False, edgecolor='k', linewidth=0.7, 
                        facecolor=bluest_color, zorder=4, snap=True)

    cbar.ax.add_patch(p)
    if remove_labels:
        ax.set_xticks([])
        ax.set_yticks([])
    else:
        ax.set_xlabel("x [pixels]", fontsize=30)
        ax.set_ylabel("y [pixels]", fontsize=30)
        ax.tick_params(axis='both', which='major', labelsize=10)
        side = "right" if cbar_side=="left" else "left"
        ax.yaxis.set_label_position(side)
        ax.yaxis.set_ticks_position(side)
    if save:
        plt.savefig(f"{save}.jpg", bbox_inches='tight')
    plt.show()

def map_pmf_plot(mag_map,label="",label_color_bar =r"$-2.5 \, \log(\mu/\langle \mu \rangle)$", vmin=-1.5, vmax=0.5,text_right_plot="Right plot",bins_limit=3,
            num_bins=100,cmap='RdYlBu',**kwargs):
    #rf"{name} {component} $\ast$ {factor} rs $\alpha = {alpha.split('_')[1]}$ mean={mean:.3f}"
    fig, (ax2, ax1) = plt.subplots(1, 2, figsize=(27, 10))
    bins = np.linspace(-bins_limit, bins_limit, num_bins + 1)
    counts, bin_edges = np.histogram(mag_map.ravel(), bins=bins)
    pmf = counts / counts.sum()
    pmf_plos = np.r_[pmf, pmf[-1]]
    mean_p = np.sum(bin_edges * pmf_plos)
    bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
    expected_pmf = sum(pmf * bin_centers)
    
    im1 = ax1.step(bin_edges,pmf_plos, alpha=0.7, linestyle="-", label=label,c="C0")
    ax1.axvline(expected_pmf,c="k",ls="--",label="Expected PMF")
    ax1.axvline(mean_p,c="r",ls="--",label="Mean")
    
    ax1.legend(loc="best", fontsize=12)
    ax1.legend(loc="best", fontsize=12)
    ax1.set_xlabel(r"$-2.5 \, \log(\mu/\langle \mu \rangle)$", fontsize=30)
    ax1.set_ylabel('Density(PMF)', fontsize=30)
    
    ax1.tick_params(axis='both', which='major', labelsize=20)
    ax1.set_xlim(-bins_limit,bins_limit)
    
    im2 = ax2.imshow(mag_map, cmap=cmap,extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    ax2.set_xticks([])
    ax2.set_yticks([])
    ax2.text(0.05, 0.95, text_right_plot, transform=ax2.transAxes, fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))
    # Create a colorbar for the mirrored plot (positioned on the left side)
    cbar2 = fig.colorbar(im2, ax=ax2, fraction=0.046, pad=0.04)
    cbar2.ax.tick_params(labelsize=30)
    cbar2.ax.set_position([ax2.get_position().x0-0.1*ax2.get_position().x0, 0.15, 0.02,  0.73]) 
    cbar2.ax.yaxis.set_ticks_position('left')
    cbar2.ax.yaxis.set_label_position('left')
    cbar2.set_label(label_color_bar, fontsize=30, labelpad=20)
    trixy = np.array([[0, 0], [1, 0], [0.5, -0.05]])
    bluest_color = im2.cmap(im2.norm(vmin))
    patch2 = Polygon(trixy, transform=cbar2.ax.transAxes, clip_on=False,edgecolor='k', linewidth=0.7, facecolor=bluest_color,zorder=4, snap=True)
    cbar2.ax.add_patch(patch2)
    plt.show()

def compare_maps_plot(mag_map_left,mag_map_right
                , vmin=-1.5, vmax=0.5,label_color_bar =r"$-2.5 \, \log(\mu/\langle \mu \rangle)$",cmap='RdYlBu',**kwargs):
    """The code assume that the two maps are normalize  but if this is not the case you can change the label_color_bar
    Args:
        mag_map_left (_type_): _description_
        mag_map_right (_type_): _description_
        text_left_plot (str, optional): _description_. Defaults to "Left plot".
        text_right_plot (str, optional): _description_. Defaults to "Right plot".
        vmin (float, optional): _description_. Defaults to -1.5.
        vmax (float, optional): _description_. Defaults to 0.5.
        label_color_bar (regexp, optional): _description_. Defaults to r"$-2.5 \, \log(\mu/\langle \mu \rangle)$".
        RdYlBu_r or RdYlBu
    """
    
    
    text_left_plot = kwargs.get("text_left_plot", "Left plot")
    text_right_plot = kwargs.get("text_right_plot","Right plot")
    
    fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(27, 10))

    im1 = ax1.imshow(mag_map_left, cmap=cmap, 
                    extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    vmin,vmax = im1.get_clim()
    ax1.set_xticks([])
    ax1.set_yticks([])
    ax1.text(0.05, 0.95,text_left_plot, transform=ax1.transAxes,
             fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))
    
    
    cbar1 = fig.colorbar(im1, ax=ax1, fraction=0.046, pad=0.04)
    cbar1.ax.tick_params(labelsize=30)
    cbar1.ax.set_position([ax1.get_position().x0-0.1*ax1.get_position().x0, 0.15, 0.02,  0.73])
    cbar1.ax.yaxis.set_ticks_position('left')
    cbar1.ax.yaxis.set_label_position('left')
    cbar1.set_label(label_color_bar, fontsize=30, labelpad=20)
    
    # Add a triangle patch to the colorbar
    bluest_color = im1.cmap(im1.norm(vmin))
    trixy = np.array([[0, 0], [1, 0], [0.5, -0.05]])
    patch1 = Polygon(trixy, transform=cbar1.ax.transAxes, clip_on=False,
                     edgecolor='k', linewidth=0.7, facecolor=bluest_color,
                     zorder=4, snap=True)
    cbar1.ax.add_patch(patch1)
    im2 = ax2.imshow(mag_map_right, cmap=cmap, 
                     extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    ax2.set_xticks([])
    ax2.set_yticks([])
    ax2.text(0.05, 0.95, text_right_plot, transform=ax2.transAxes,
             fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))
    cbar2 = fig.colorbar(im2, ax=ax2, fraction=0.046, pad=0.04)
    cbar2.ax.tick_params(labelsize=30)
    cbar2.ax.set_position([ax2.get_position().x1+0.005*ax2.get_position().x1, 0.15, 0.05, 0.73]) 
    cbar2.ax.yaxis.set_ticks_position('right')
    cbar2.ax.yaxis.set_label_position('right')
    cbar2.set_label(label_color_bar, fontsize=30, labelpad=20)
    
    # Add a similar patch to the second colorbar
    patch2 = Polygon(trixy, transform=cbar2.ax.transAxes, clip_on=False,
                      edgecolor='k', linewidth=0.7, facecolor=bluest_color,
                      zorder=4, snap=True)
    cbar2.ax.add_patch(patch2)
    plt.show()
    
    
def compare_maps_pmf(mag_map_left,mag_map_right
                , vmin=-1.5, vmax=0.5,label_color_bar =r"$-2.5 \, \log(\mu/\langle \mu \rangle)$",bins_limit=3,num_bins=100,cmap="RdYlBu",**kwargs):
    """The code assume that the two maps are normalize  but if this is not the case you can change the label_color_bar
    Args:
        mag_map_left (_type_): _description_
        mag_map_right (_type_): _description_
        text_left_plot (str, optional): _description_. Defaults to "Left plot".
        text_right_plot (str, optional): _description_. Defaults to "Right plot".
        vmin (float, optional): _description_. Defaults to -1.5.
        vmax (float, optional): _description_. Defaults to 0.5.
        label_color_bar (regexp, optional): _description_. Defaults to r"$-2.5 \, \log(\mu/\langle \mu \rangle)$".
        'RdYlBu_r' or 'RdYlBu_r'
    """
    #, text_left_plot = "Left plot", text_right_plot = "Right plot"
    text_left_plot = kwargs.get("text_left_plot", "Left plot")
    text_right_plot = kwargs.get("text_right_plot","Right plot")
    fig, (ax1,ax3,ax2) = plt.subplots(1, 3, figsize=(47, 10))

    im1 = ax1.imshow(mag_map_left, cmap=cmap, 
                    extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    ax1.set_xticks([])
    ax1.set_yticks([])
    ax1.text(0.05, 0.95,text_left_plot, transform=ax1.transAxes,
             fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))
    
  
    cbar1 = fig.colorbar(im1, ax=ax1, fraction=0.046, pad=0.04)
    cbar1.ax.tick_params(labelsize=20)
    cbar1.ax.set_position([ax1.get_position().x0-0.1*ax1.get_position().x0, 0.15, 0.02,  0.73])
    cbar1.ax.yaxis.set_ticks_position('left')
    cbar1.ax.yaxis.set_label_position('left')
    cbar1.set_label(label_color_bar, fontsize=30, labelpad=20)
    
    # Add a triangle patch to the colorbar
    bluest_color = im1.cmap(im1.norm(vmin))
    trixy = np.array([[0, 0], [1, 0], [0.5, -0.05]])
    patch1 = Polygon(trixy, transform=cbar1.ax.transAxes, clip_on=False,
                     edgecolor='k', linewidth=0.7, facecolor=bluest_color,
                     zorder=4, snap=True)
    cbar1.ax.add_patch(patch1)
    # ---------------------------
    # Right: Mirrored version
    # ---------------------------
    # Create the horizontally mirrored data using np.fliplr
    #mag_map_2d_norm_mirror = np.fliplr(mag_map_2d_norm)
    im2 = ax2.imshow(mag_map_right, cmap=cmap, 
                     extent=[-2, 2, -2, 2], vmin=vmin, vmax=vmax)
    ax2.set_xticks([])
    ax2.set_yticks([])
    ax2.text(0.05, 0.95, text_right_plot, transform=ax2.transAxes,
             fontsize=30, fontweight='bold', va='top', ha='left',bbox=dict(facecolor='white', edgecolor='none', alpha=0.8))
    # Create a colorbar for the mirrored plot (positioned on the left side)
    cbar2 = fig.colorbar(im2, ax=ax2, fraction=0.046, pad=0.04)
    cbar2.ax.tick_params(labelsize=20)
    cbar2.ax.set_position([ax2.get_position().x1+0.005*ax2.get_position().x1, 0.15, 0.05, 0.73]) 
    cbar2.ax.yaxis.set_ticks_position('right')
    cbar2.ax.yaxis.set_label_position('right')
    cbar2.set_label(label_color_bar, fontsize=30, labelpad=20)
    
    # Add a similar patch to the second colorbar
    patch2 = Polygon(trixy, transform=cbar2.ax.transAxes, clip_on=False,
                      edgecolor='k', linewidth=0.7, facecolor=bluest_color,
                      zorder=4, snap=True)
    cbar2.ax.add_patch(patch2)
    colors_a = {text_left_plot:"C0",text_right_plot:"C1"}
    colors_b = {text_left_plot:"k",text_right_plot:"r"}
    for key,map in {text_left_plot:mag_map_left,text_right_plot:mag_map_right}.items():
        bins = np.linspace(-bins_limit, bins_limit, num_bins + 1)
        counts, bin_edges = np.histogram(map.ravel(), bins=bins)
        pmf = counts / counts.sum()
        pmf_plos = np.r_[pmf, pmf[-1]]
        mean_p = np.sum(bin_edges * pmf_plos)
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2
        expected_pmf = sum(pmf * bin_centers)
        ax3.step(bin_edges,np.r_[pmf, pmf[-1]], alpha=0.7, linestyle="-", label=f"PMF ({key})")
        #ax3.axvline(expected_pmf,ls="--",label=f"Expected pmf {key}",alpha=0.7,c=colors_a[key])
        #ax3.axvline(mean_p,ls="-",label=f"mean_p {key}",alpha=0.7,c=colors_b[key])
    
    ax3.legend(loc="best", fontsize=21)
    ax3.set_xlabel(r"$-2.5 \, \log(\mu/\langle \mu \rangle)$", fontsize=30)
    ax3.set_ylabel('Density(PMF)', fontsize=30)
    ax3.tick_params(axis='both', which='major', labelsize=20)
    ax3.set_xlim(-bins_limit, bins_limit)
    
    plt.show()
    
# def plot_mag_map_pmf(mag_map_dic,key_map='mag_map_2d_norm_conv', title="", save="", cbar_label="", show=False, text=None):
#     image = ""  # Image name (for labeling)
    
#     # Define GridSpec with custom subplot sizes
#     fig = plt.figure(figsize=(30, 15))
#     if save:
#         plt.ioff()
#     gs = gridspec.GridSpec(2, 2, width_ratios=[1, 1], height_ratios=[1, 1])  # Add space for text
#     # Set general title for the figure
#     fig.suptitle(f"{title}", fontsize=25)
    
#     # First plot (n == 0): Image with colorbar
#     ax0 = fig.add_subplot(gs[:, 0])  # Span both rows for larger size
#     im = ax0.imshow(mag_map_dic[key_map], cmap='RdYlBu')
#     divider = make_axes_locatable(ax0)
#     cax = divider.append_axes("right", size="5%", pad=0.1)  # Adjust size and padding
#     ax0.set_xlabel("x [pixels]", fontsize=20)
#     ax0.set_ylabel("y [pixels]", fontsize=20)
#     ax0.tick_params(axis='both', which='major', labelsize=20)
#     cbar = plt.colorbar(im, cax=cax)
#     cbar.ax.tick_params(labelsize=20)
#     cbar.set_label(cbar_label, fontsize=20)
    
#     # Second plot (n == 1): Histogram with PMF and PDF
#     ax1 = fig.add_subplot(gs[1, 1])  # Small, symmetric plot
    
#     ax1.step(
#         mag_map_dic["bin_edges"], np.r_[mag_map_dic["pmf"], mag_map_dic["pmf"][-1]],
#         alpha=0.5, color='blue') 
#         #label=rf"$|mean| = {mag_map_dic['expected_pmf_abs']:.2f}, \quad |std| = {np.sqrt(mag_map_dic['var_pmf_abs']):.2f}$"
#     #)
#     ax1.set_xlabel(r"$-2.5 \, \log(\mu/\langle \mu \rangle)$", fontsize=16)
#     ax1.set_ylabel('Density (PMF)', fontsize=16)
#     ax1.set_xlim(-3, 3)
#     ax1.legend(loc="best", fontsize=14)
#     ax1.tick_params(axis='both', which='major', labelsize=16)
    
#     # Move y-axis of the second plot to the right
#     ax1.yaxis.set_label_position("right")
#     ax1.yaxis.tick_right()
    
#     # Text area below the small plot (if text is provided)
#     if text:
#         ax_text = fig.add_subplot(gs[1, 1])  # Use the bottom-right quadrant
#         ax_text.axis('off')  # Turn off the axes
#         ax_text.text(0.5, 0.5, text, fontsize=16, ha='center', va='center', wrap=True)
    
#     # Adjust layout
#     plt.tight_layout(rect=[0, 0, 1, 0.95])
    
#     if save:
#         plt.savefig(f"{save}.png")
#         plt.close()
#     else:
#         plt.show()
    
#     return



        
        
# def mag_map_plot(map,main_title=None,titles=None,vmax=None,vmin=None,color_bar_label =r"$-2.5 \, \log(\mu/\langle \mu \rangle)$",save=""):
#     #-0.32 is a value that Evencio askme for 
#     #mag_map_2d_norm_conv
#     if isinstance(map,list):
#         map = np.stack(map)
#         n_plots=map.shape[0]
#     else:
#         map = np.stack([map])
#         n_plots=map.shape[0]
#     titles = titles or [""]*n_plots
#     main_title = main_title or ""
#     fig, axes = plt.subplots(1, n_plots, figsize=(20*n_plots, 10*n_plots))
#     axes = np.atleast_1d(axes)
#     axes = axes.flat
#     for i,ax in enumerate(axes):
#         im = ax.imshow(map[i],vmax=vmax,cmap='RdYlBu')
#         divider = make_axes_locatable(ax)
#         cax = divider.append_axes("right", size="5%", pad=0.1)  # Adjust size and padding
#         ax.set_xlabel("x [pixels]", fontsize=30)
#         ax.set_ylabel("y [pixels]", fontsize=30)
#         ax.tick_params(axis='both', which='major', labelsize=30)
#         ax.set_title(titles[i], fontsize=30)
#         cbar = plt.colorbar(im, cax=cax)
#         cbar.ax.tick_params(labelsize=30)
#         cbar.set_label(color_bar_label, fontsize=30)
    
#     # Adjust tight_layout to leave space for the suptitle
#     plt.tight_layout()  # Reserve top 5% for the suptitle
#     #fig.suptitle(main_title, fontsize=25,y=0.7)
#     if save:
#         plt.savefig(f"plot_{save}.jpg")
#         plt.close()
#     else:
#         plt.show()
        


 
    


    
    
# def hist_n_image(dic,name,save="",images: list=None):
#     #name = "2M1310-1714"
#     title = name 
#     system = dic[name]
#     if not images:
#         images = system.keys()
#     fig, axes = plt.subplots(1, 1, figsize=(20, 10))
#     #plt.ioff()
#         # Plot the three bar series
#     yvalue = "pmf"
#     color = ["C0","C1","C7","C3","C4","C5"]
#     [axes.step(system[image]["bin_edges"],np.r_[system[image][yvalue], system[image][yvalue][-1]], alpha=0.5,color=color[n], label=rf"${image} \quad |mean| = {system[image]['expected_pmf_abs']:.2f}, \quad |std| = {np.sqrt(system[image]['var_pmf_abs']):.2f}$") for n,image in enumerate(images)]
#     #[ax.text(ax.get_xlim()[0]+ax.get_xlim()[1]*0.2, ax.get_ylim()[0]+ax.get_ylim()[1]*(0.8-n/10), rf"$m_1(\mu) = {images_[title][image]['mean_magnification']:.2f}, \quad m_2(\sigma^2) = {images_[title][image]['variance']:.2f}$", color='k', fontsize=10)  for n,image in enumerate(images_[title].keys())] #, bbox=dict(facecolor='white', alpha=0.7)]
#     axes.set_title(f'{title}({results[results.name==title].mass_models.values[0]})', fontsize=14)
#     # Optional: set x-label, y-label for each subplot
#     axes.set_xlabel(r"$-2.5 \, \log(\mu/\langle \mu \rangle)$", fontsize=12)
#     axes.set_ylabel(f'Density({yvalue})', fontsize=12)

#     # Add legend
#     axes.legend(loc="best", fontsize=12)

#     # Make sure tick labels have the same size
#     axes.tick_params(axis='both', which='major', labelsize=20)
#     axes.set_xlim(-3,3)
#     if save:
#         plt.savefig(f"plot_{name}.jpg")
#         plt.close()
#     else:
#         plt.show()


# def plot_mag_map(array, save="", show=False,cbar_label="",title =""):
#     fig, ax = plt.subplots(figsize=(20, 20))
#     im = ax.imshow(array,cmap='RdYlBu')
#     # Create a divider for the axes to append the colorbar
#     divider = make_axes_locatable(ax)
#     cax = divider.append_axes("right", size="5%", pad=0.1)  # Adjust size and padding here
#     ax.set_xlabel("x [pixels]", fontsize=20)
#     ax.set_ylabel("y [pixels]", fontsize=20)
#     ax.tick_params(axis='both', which='major', labelsize=20)
#     ax.set_title(title, fontsize=30)
#     # Add the color bar
#     # Add the color bar with fontsize
#     cbar = plt.colorbar(im, cax=cax)
#     cbar.ax.tick_params(labelsize=20)
#     if cbar_label:
#         cbar.set_label(cbar_label, fontsize=20)
#     plt.tight_layout()
#     if save:
#         plt.savefig(f"{save}.png")
#     if show:
#         plt.show()
#     plt.close()


# Function to plot magnification map and PMF


def array_to_latex_table(array,headers=None):
    """
    Converts a 2D array (list of lists) into a LaTeX table.
    If the first column values contain the prefix "alpha_", it is removed.
    A header row is generated dynamically based on the number of columns.
    
    Parameters:
        array (list of lists): 2D array to convert into a LaTeX table.
        
    Returns:
        None: Prints the LaTeX table code.
    """
    # Check if the array is empty or has no columns
   # if not array or not array[0]:
    #    print("Empty or invalid array provided.")
     #   return
    
    n_cols = len(array[0])
    
    # Build the header row dynamically: first column header remains blank.
    if headers is None:
        headers = [""] + [f"Col{j+1}" for j in range(1, n_cols)]
    else:
        headers = [""] + headers
    header_line = " & ".join(headers) + " \\\\\n"
    
    # Start building the LaTeX table string
    latex_str = "\\begin{table}\n"
    latex_str += "    \\centering\n"
    #latex_str += "renewcommand{\arraystretch}{1.2}"
    latex_str += "    \\begin{tabular}{" + "c" * n_cols + "}\n"
    latex_str += "        \\hline\n"
    # Add the dynamically created header row
    latex_str += "        " + header_line
    latex_str += "        \\hline\n"
    
    # Process each row in the array
    for row in array:
        # Process the first element: remove "alpha_" if present.
        first_item = row[0]
        if isinstance(first_item, str) and first_item.startswith("alpha_"):
            first_item = first_item.replace("alpha_", "")
        # Convert each item in the row to a string (first element processed separately)
        row_items = [str(first_item)] + [str(item) for item in row[1:]]
        # Join them with " & " and append LaTeX newline
        row_line = " & ".join(row_items) + " \\\\\n"
        latex_str += "        " + row_line
    
    # Finish table with a bottom horizontal line and caption/label
    latex_str += "        \\hline\n"
    latex_str += "    \\end{tabular}\n"
    latex_str += "    \\caption{Generated table with dynamic headers}\n"
    latex_str += "    \\label{tab:generated}\n"
    latex_str += "\\end{table}"
    
    print(latex_str)


import numpy as np
from lmfit import Model
from scipy.integrate import quad
import warnings

import multiprocessing
from functools import partial

warnings.simplefilter(action='ignore', category=FutureWarning)

def linear_func(x, slope, intercept):
        return slope * x + intercept

def linear_model(x,y):
    linear_fit_model = Model(linear_func)
    params = linear_fit_model.make_params(slope=1, intercept=0)
    result = linear_fit_model.fit(y,x=x, params=params)
    slope_fit, intercept_fit = result.params['slope'].value, result.params['intercept'].value
    return slope_fit, intercept_fit


#    for key, y_noisy in y_noisy_list.items():
#         y_noisy_between_barriers = np.concatenate((y_noisy[(x >= x_barrier1) & (x <= x_barrier2)], y_noisy[(x >= x_barrier3) & (x <= x_barrier4)]))
#         ax1_1.scatter(x_between_barriers, y_noisy_between_barriers, alpha=1, label=f'Data (between limits) imagen {key}', zorder=3) #pensar en colores
#         ax1_1.plot(X,Y[key],label=f'spectrum imagen {key}', alpha=0.6, zorder=1)
#         # Fit a single linear model to the data between the barriers
#         linear_fit_model = Model(linear_fit)
#         params = linear_fit_model.make_params(slope=1, intercept=0)
#         result = linear_fit_model.fit(y_noisy_between_barriers, x=x_between_barriers, params=params)
#         slope_fit, intercept_fit = result.params['slope'].value, result.params['intercept'].value
#         ax1_1.plot(x_between_barriers, linear_fit(x_between_barriers, slope_fit, intercept_fit), label=f'Fitted Linear Function for {key}', color='r', zorder=4)
        
#         # Calculate the area under the linear function between Barrier 1 and Barrier 4
#         area, _ = quad(linear_fit, x_barrier1, x_barrier4, args=(slope_fit, intercept_fit))

#         # Calculate the curve after subtracting the area obtained from the fit
#         y_curve = y_noisy - linear_fit(x, slope_fit, intercept_fit)
#         Y_curve = Y[key] - linear_fit(X, slope_fit, intercept_fit)
#         suma = np.sum(y_curve[(x_barrier5 <= x) & (x_barrier6 >= x)])

#         # Plot the curve resulting from subtracting the area
        
#         ax1_2.plot(X, Y_curve, label=f'Line without continium {key}', color='grey',alpha=0.5)
#         ax1_2.hlines(y=[0, 0], xmin=min(x), xmax=max(x), colors='k', linestyles='dashed', zorder=3)

def bootstrap_single_iteration(x, y, _):
    pto_medio=np.min(x)+(np.max(x)-np.min(x))/2
    if isinstance(x,np.ndarray):
        x_L,y_L=x[x<pto_medio],y[x<pto_medio]
        x_R,y_R=x[x>pto_medio],y[x>pto_medio]
    else:
        x_L,y_L=x.values[x<pto_medio],y.values[x<pto_medio]
        x_R,y_R=x.values[x>pto_medio],y.values[x>pto_medio]
    # Resample the dataset with replacement
    resampled_indices_R = np.random.choice(len(x_R), len(x_R), replace=True)
    resampled_x_R = x_R[resampled_indices_R]
    resampled_y_R = y_R[resampled_indices_R]
   
    resampled_indices_L = np.random.choice(len(x_L), len(x_L), replace=True)

    resampled_x_L = x_L[resampled_indices_L]
    resampled_y_L = y_L[resampled_indices_L]
    
    model = Model(linear_fit)
    params = model.make_params(slope=1, intercept=0)
    resampled_x=np.concatenate((resampled_x_L, resampled_x_R))
    
    resampled_y=np.concatenate((resampled_y_L, resampled_y_R))
    result = model.fit(resampled_y, params, x=resampled_x)

    fitted_m = result.params["slope"].value
    fitted_b = result.params["intercept"].value

    
    x_min = min(resampled_x)
    x_max = max(resampled_x)

    area = (fitted_m * (x_max**2 / 2) + fitted_b * x_max) - (fitted_m * (x_min**2 / 2) + fitted_b * x_min)
    return fitted_m, fitted_b, area

def bootstrap_Area(x, y, n_bootstrap=5000):
    ncpu = multiprocessing.cpu_count()
    bootstrap_func = partial(bootstrap_single_iteration, x, y)

    with multiprocessing.Pool(ncpu) as pool:
        results = pool.map(bootstrap_func, range(n_bootstrap))

    slope_fit_estimates, intercept_estimates, area_estimates = zip(*results)
    mean_slope_fit_estimates,std_slope_fit_estimates = np.median(slope_fit_estimates),np.std(slope_fit_estimates)
    mean_intercept_estimates,std_intercept_estimates = np.median(intercept_estimates),np.std(intercept_estimates)
    mean_area,std_area = np.median(area_estimates),np.std(area_estimates)
    
    
    return mean_slope_fit_estimates,std_slope_fit_estimates,mean_intercept_estimates,std_intercept_estimates,mean_area,std_area, slope_fit_estimates, intercept_estimates, area_estimates


import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button,RangeSlider
from lmfit import Model
from scipy.integrate import quad
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import pandas as pd
import os
import multiprocessing
from functools import partial
from .utils import format_scientific
from screeninfo import get_monitors
from .functions  import linear_fit




def update(Lp,Lr,Wslider_lc,Wslider_core,Wslider_rc,Wslider_Lp,Wslider_Rp,Fslider_Lp,Fslider_Rp, Y):
    
    Wslider_Lp_min, Wslider_Lp_max = Wslider_Lp.val[0], Wslider_Lp.val[1]
    Lp.set_xlim(Wslider_Lp_min,Wslider_Lp_max)
    plt.draw()


    
#     global slider_barrier12,barrier_values,y_curve,y_noisy,x_between_barriers,y_noisy_between_barriers,range_values  #barrier_values, y_curve, y_noisy_between_barriers
#     # Extract barrier positions from sliders
#     x_barrier1, x_barrier2 = slider_barrier12.val[0], slider_barrier12.val[1]
#     x_barrier3, x_barrier4 = slider_barrier34.val[0], slider_barrier34.val[1]
#     x_barrier5, x_barrier6 = slider_barrier56.val[0], slider_barrier56.val[1]

#     # Additional barrier positions
#     x_slider_range, y_slider_range = slider_range.val, slider_range_y.val
#     x_slider_range2, y_slider_range2 = slider_range2.val, slider_range_y2.val

#     # Combine the data between barriers 1 and 2, and between barriers 3 and 4
#     barrier_values = [x_barrier1, x_barrier2, x_barrier3, x_barrier4, x_barrier5, x_barrier6]
#     range_values = [ *slider_range.val, *slider_range_y.val,*slider_range2.val, *slider_range_y2.val]
#     x_between_barriers = np.concatenate((x[(x >= x_barrier1) & (x <= x_barrier2)], x[(x >= x_barrier3) & (x <= x_barrier4)]))
    
#     ax1_1.clear()
#     ax1_2.clear()
    
#     for key, y_noisy in y_noisy_list.items():
#         y_noisy_between_barriers = np.concatenate((y_noisy[(x >= x_barrier1) & (x <= x_barrier2)], y_noisy[(x >= x_barrier3) & (x <= x_barrier4)]))
#         ax1_1.scatter(x_between_barriers, y_noisy_between_barriers, alpha=1, label=f'Data (between limits) imagen {key}', zorder=3) #pensar en colores
#         ax1_1.plot(X,Y[key],label=f'spectrum imagen {key}', alpha=0.6, zorder=1)
#         # Fit a single linear model to the data between the barriers
#         linear_fit_model = Model(linear_fit)
#         params = linear_fit_model.make_params(slope=1, intercept=0)
#         result = linear_fit_model.fit(y_noisy_between_barriers, x=x_between_barriers, params=params)
#         slope_fit, intercept_fit = result.params['slope'].value, result.params['intercept'].value
#         ax1_1.plot(x_between_barriers, linear_fit(x_between_barriers, slope_fit, intercept_fit), label=f'Fitted Linear Function for {key}', color='r', zorder=4)
        
#         # Calculate the area under the linear function between Barrier 1 and Barrier 4
#         area, _ = quad(linear_fit, x_barrier1, x_barrier4, args=(slope_fit, intercept_fit))

#         # Calculate the curve after subtracting the area obtained from the fit
#         y_curve = y_noisy - linear_fit(x, slope_fit, intercept_fit)
#         Y_curve = Y[key] - linear_fit(X, slope_fit, intercept_fit)
#         suma = np.sum(y_curve[(x_barrier5 <= x) & (x_barrier6 >= x)])

#         # Plot the curve resulting from subtracting the area
        
#         ax1_2.plot(X, Y_curve, label=f'Line without continium {key}', color='grey',alpha=0.5)
#         ax1_2.hlines(y=[0, 0], xmin=min(x), xmax=max(x), colors='k', linestyles='dashed', zorder=3)
#         # Display the area as text inside the plot
#     # Plot the fitted linear function   
    
#     ax1_1.axvline(x=line_info["line_center"], color='k', label='Line',linestyle ="--")
#     ax1_1.fill_betweenx([-100,100], x_barrier1, x_barrier2,
#                         label='left continium', color='purple', alpha=0.2)

#     # Create a region between x_barrier3 and x_barrier4
#     ax1_1.fill_betweenx([-100,100], x_barrier3, x_barrier4,
#                         label='right continium', color='g', alpha=0.2)

# # Rest of the code remains the same up to this point

#     ax1_1.legend(fontsize='xx-small', frameon=False,ncol=3)
#     ax1_1.set_xlim(x_slider_range[0], x_slider_range[1])
#     ax1_1.set_ylim(y_slider_range[0], y_slider_range[1])
#     ax1_1.set_xlabel('wavelength')
#     ax1_1.set_ylabel('Flux')
    
#     ax1_2.fill_between(x[(x_barrier5 <= x) & (x_barrier6 >= x)], 
#                    -100,100, 
#                    label='line core region', color='r', alpha=0.3, zorder=2)
#     #ax1_2.axvline(x=x_barrier5, color='m', linestyle='--', label='Barrier 5')
#     #ax1_2.axvline(x=x_barrier6, color='c', linestyle='--', label='Barrier 6')
#     ax1_2.axvline(x=line_info["line_center"], color='k', label='Line',linestyle ="--")
#     ax1_2.set_xlim(x_slider_range2[0], x_slider_range2[1])
#     ax1_2.set_ylim(y_slider_range2[0], y_slider_range2[1])
#     ax1_2.set_xlabel('wavelenght')
#     ax1_2.set_ylabel('Flux')
#     ax1_2.legend(fontsize='small', frameon=False,ncol=3)
    
#     plt.draw()


import os
from pathlib import Path
import numpy as np
from scipy.spatial import distance
import json
import stat
import subprocess
from .utils import columns_to_float
import pandas as pd
from .mass_models import mass_models

# Get the path to the current file
module_dir = Path(__file__).resolve().parent

# Navigate to the root of the qumas package (i.e., one level up from LensmodelWrapper)
qumas_root = module_dir.parent

# Define the path to the CSV file (e.g., in the data/ subdirectory)
csv_path = qumas_root / "Tables" / "filter_unique.csv"
class lensmodel_handler:
    """modeling_path: where will go the results of our modeling
    system:and pandas data frame from our census with all information to model our system
    center_mass_error
    relative_flux_error
    number_of_lens
    astrometry_error
    use_real_error_flux
    R_chi"""
    
    
    def __init__(self, modeling_path:str,system:pd.DataFrame
                 ,json_models=f"{module_dir}/models"): 
        self.model_setup_path = json_models
        self.modeling_path = modeling_path
        self.system = system
        self.system_name = self.system.name.values[0]
        self.n_images = int(self.system.model_images.values[0]) #images can be modeled
        self.mass_distribution = "SIE"
        ###set_up_data#####
        self.number_of_lens = 1
        self.zl,self.zs=[i for i in self.system[["z_l","z_s"]].values[0]]
        self.available_bands = self.system.available_bands.values[0]
        self.band_to_model = self.available_bands[0]
        self.images,self.lens = columns_to_float(self.system,self.band_to_model)
        #self.models_file = "double_models.json" if len(self.images)==2 else "quad_models.json"
        self.models_file = "quad_models.json" 
        with open(os.path.join(self.model_setup_path,self.models_file), 'r') as file_mass:
             self.mass_models = json.load(file_mass)
        #prior Er
        self.p1 = round(np.max(distance.pdist(np.array([self.images["RA"].values,self.images["DEC"].values]).T,metric='euclidean')/2),2)
        #if is not create yet create directories for the results and a directory in where will run the code.
        if modeling_path != None:
            self.data_path = os.path.join(self.modeling_path,f"{self.system_name}.dat")
            if not os.path.isdir(self.modeling_path):
                os.mkdir(self.modeling_path)
            #if os.path.isdir(os.path.join(self.modeling_path,"final_result"))==False:
             #   os.mkdir(os.path.join(self.modeling_path,"final_result"))
        else:
            self.data_path = None
    def __str__(self):
        return f"set up modeling for {self.system_name}"
    
    #def __call__(self):

    #maybe if moment to use **kwargs?
    def data_writter(self,use_real_error_flux=False,use_real_astrometry_error=False,free_center_of_mass=False, \
                    astrometry_error=0.003,center_mass_error=0.003,relative_flux_error=0.2,band_to_model=None,use_informed_flux=False):
        band_to_model = band_to_model if band_to_model in self.available_bands else self.band_to_model
        images, lens = columns_to_float(self.system, band_to_model) if band_to_model else (self.images, self.lens)
        #how can handle system with more image that the ones i want to use in the first model ?
        images_lens_centered = images[["RA","DEC"]].values-lens[["RA","DEC"]].values[0]
        lens_lens_centered = lens[["RA","DEC"]].values-lens[["RA","DEC"]].values[0]
        error_lens_lens_centered = np.sqrt(lens[["dRA","dDEC"]].values**2+lens[["dRA","dDEC"]].values[0]**2)
        magnitudes = images[band_to_model].values
        band_info = self.system.loc[images[band_to_model].index][["photometric_system","Telescope","instrument"]].values[0]
        if not isinstance(band_info[0],str):
            band_info[0] = "Vega"
        mask = ((filter_info["photometric_system"] == "Vega" if "v"  in band_info[0] else filter_info["photometric_system"] == band_info[0]) & (filter_info["Telescope"] == band_info[1]) &(filter_info["instrument"] == band_info[2]) & 
                (filter_info["band_to_model"] == band_to_model.replace("band_", "")))
        
        band_matching_rows = filter_info[mask]
        if len(band_matching_rows) == 0:
            print('no band_info specific band info for this case')
            band_info = {i: None for i in filter_info.columns}
            band_info["band_to_model"] = band_to_model.replace("band_", "")
        else:
            band_info = band_matching_rows.iloc[0].to_dict()
        #print(band_matching_rows.iloc[0].to_dict())
        if use_informed_flux:
            flux = magnitudes
        else:
            flux=np.array([((-1)**(ii))*10**(((-magnitudes[ii])+min(magnitudes))/2.5) for ii in range(self.n_images)]) # divided flux
        flux_error = relative_flux_error*np.abs(flux)
        if use_real_error_flux:
            flux_error = images[band_to_model.replace("band",'error')].values
            #return print("Not define yet a formula for the propagation of the error in the flux")
            #flux_error = image[2] #error propagation from flux
        astrometry_error = self.n_images*[astrometry_error]
        if use_real_astrometry_error:
            error_images_lens_centered = np.sqrt(images[["dRA","dDEC"]].values**2+lens[["dRA","dDEC"]].values[0]**2)
            #could be the max image error or the error in the specific image
            astrometry_error=np.max(error_images_lens_centered,axis=1)
            astrometry_error = self.n_images*[np.max(error_images_lens_centered.flatten())]
        number_of_lens = self.number_of_lens
        
        no_lens = True if number_of_lens<1 else False
        if free_center_of_mass:
            #what is the idea behind this ? maybe take the center of mass
            #between the two images ?
            # x1,x2=images_lens_centered
            # y1,y2=images_lens_centered
            # p1=np.array([x1,y1])
            # p2=np.array([x2,y2])
            # p3=np.array([0,0])
            # error_cm=[np.linalg.norm(np.cross(p2-p1, p1-p3))/np.linalg.norm(p2-p1)]
            if len(images_lens_centered)==2:
                return print("not yet")
            if len(images_lens_centered)>2:
                return print("not yet")
        self.lens_lens_centered = lens_lens_centered
        if "-" in self.mass_distribution:
            number_of_lens = int(self.mass_distribution.split("-")[1][0])
            #print("number_of_lens",number_of_lens)
        if no_lens:
            return print("not yet")
        if isinstance(center_mass_error, float):
            center_mass_error = [center_mass_error]
            if number_of_lens>1:
                center_mass_error+= [0.5]*(number_of_lens-1)
        time_delay,time_delay_error = 0,1000
        if self.data_path:
            f = open(os.path.join(self.data_path), 'w')
            f.write(f"#The system name is: {self.system_name} \n")
            f.write(f"{number_of_lens} # number of lens galaxy\n")
            for ii in range(number_of_lens):
                ra,dec = lens_lens_centered[ii]
                f.write(f"{ra:.3f} {dec:.3f} {center_mass_error[ii]:.3f} {center_mass_error[ii]:.3f} #position \n")
                f.write(f"0.0 1000. # R_eff observed: 0.59 +/- 0.06 \n")
                f.write(f"0.0 1000. # PA unconstrained in observations \n")
                f.write(f"0.0 1000. # observed e < 0.07 at 1sigma \n")
            f.write(f"\n")
            f.write(f"1 # nºsource \n")
            f.write(f"{self.n_images} # n° images of the source \n")
            f.write(f"#band used {band_to_model} \n \n")
            f.write("#Ra Dec flux sigmax sigmaflux tdel sigma(tdel) part "+'\n') 
            for n,image in enumerate(images[["component",band_to_model,f"error_{band_to_model[5:]}"]].values): #peligroso
                ra,dec=images_lens_centered[n]
                #flux=((-1)**(ii))*10**(((-image[1])+min(magnitudes))/2.5) # divided flux
                #if np.max(image[1])<=5:
                #   flux=((1)**(ii))*image[1]
                #if np.max(image[1])>30:
                #   flux = ((-1)**(ii))*image[1]/np.max(image[1])
                f.write(f"{ra:.3f} {dec:.3f} {flux[n]:.3f} {astrometry_error[n]:.3f} {flux_error[n]:.3f} {time_delay:.3f} {time_delay_error:.3f} {image[0]} \n")
            f.close()
        data_model = {"name":self.system_name,"component":images["component"].values,"ra":images_lens_centered[:,0],"dec":images_lens_centered[:,1],"astrometry_error":astrometry_error \
                    ,"flux":flux,"flux_error":flux_error,"magnitudes":magnitudes,"band_to_model":band_to_model,"lens_ra":lens_lens_centered[:,0],"lens_dec":lens_lens_centered[:,1] \
                    ,"center_mass_error":center_mass_error,"zl":self.zl,"zs":self.zs}
        data_model.update(band_info)
        return data_model
    def model_writter(self,mass_distribution=None,name_run ="model_run",kwards_special={}):
        with open(os.path.join(self.model_setup_path,"model_setup.json"), 'r') as file:
            setup_model = json.load(file)
        if mass_distribution==None:
            mass_distribution = self.mass_distribution
        number_of_lens = self.number_of_lens
        if "-" in self.mass_distribution:
            number_of_lens = int(self.mass_distribution.split("-")[1][0])
            mass_distribution = self.mass_distribution.split("-")[0]
        if mass_distribution not in self.mass_models.keys():
                return print(f"{mass_distribution} mass model is not available")
        mass = self.mass_models[mass_distribution]
        if self.data_path != None:
            f = open(os.path.join(self.modeling_path,f"{name_run}.dat"), 'w')
            f.write(f"#mass_distribution = {mass_distribution} \n")
            #this should be go to data_model
            kwards_basic = {"set_data":self.data_path,'set_zlens' :self.zl,'set_zsrc' :self.zs,"set_lens_number": number_of_lens}
            for keys in kwards_basic.keys(): setup_model["lensmodel_values"][keys] =kwards_basic[keys]
            for keys in kwards_special.keys(): setup_model["lensmodel_values"][keys] =kwards_special[keys]
            for key in setup_model["lensmodel_values"]: f.write(f"{setup_model['lensmodel_keys'][key]} {setup_model['lensmodel_values'][key]} {setup_model['lensmodel_coments'][key]}")
            f.write(mass_models(mass_distribution,number_of_lens, self.p1, self.lens_lens_centered, self.modeling_path))
            #for key in mass: f.write(f"{mass[key].replace('p1',str(self.p1)).replace('path',self.modeling_path)}")
            f.close()
        return
    
    def writter_model_data(self,mass_distribution=None, name_run="model_run",**kwards):
        #mass_distribution
        #if mass_distribution==None:
        #print(mass_distribution)
        self.mass_distribution = mass_distribution or self.mass_distribution 
        data_model = self.data_writter(**kwards)
        self.model_writter(mass_distribution=mass_distribution,name_run =name_run)
        data_model["name_run"]= name_run
        data_model["mass_distribution"]= mass_distribution
        return data_model


"""Main `qumas` CLI."""

import os
import sys

import click

from qumas import __version__

this_file_location = os.path.dirname(os.path.realpath(os.path.abspath(__file__)))


def version_msg():
    """qumas version, location and Python version.

    Get message about qumas version, location
    and Python version.
    """
    python_version = sys.version[:3]
    message = u"qumas %(version)s from {} (Python {})"
    location = os.path.dirname(this_file_location)
    return message.format(location, python_version)


@click.command(context_settings=dict(help_option_names=[u'-h', u'--help']))
@click.version_option(__version__, u'-V', u'--version', message=version_msg())
# @click.option('-v', '--verbose', is_flag=True, help='Print debug information', default=False)
def main(
    # verbose,
):
    """TODO Write this content that gets renders when invoking with --help flag!

    Eg:
    Create a Project from the project template.

    Cookiecutter Python Package is Free/Libre Open Source Software. If you would
    like to get in touch, please see
    https://github.com/boromir674/cookiecutter-python-package.
    """
    try:
        pass
    except Exception as error:  # pylint: disable=broad-except
        click.echo(error)
        sys.exit(1)


if __name__ == "__main__":  # pragma: no cover
    main()
